# This test case is for plate data plot3d file
# TODO: Add the docstring test

import unittest
import doctest
import numpy as np
from src.io import GridIO, FlowIO


class TestIO(unittest.TestCase):
    def test_grid_io(self):

        # Test with the plate data
        grid = GridIO('../data/plate_data/plate.sp.x')
        # Print the doc string
        print(grid)
        grid.read_grid()

        # Simple assertion to test the import
        self.assertEqual(grid.grd.shape, (grid.ni, grid.nj, grid.nk, 3, 1))

        return

    def test_flow_io(self):

        # Test with the plate data
        flow = FlowIO('../data/plate_data/sol-0000010.q')
        # Print the doc string
        print(flow)
        flow.read_flow()

        # Simple assertion to test the import
        self.assertEqual(flow.q.shape, (flow.ni, flow.nj, flow.nk, 5, 1))

    def test_multi_block(self):
        """
        This tests the basic slicing functionality in the code
        Returns:

        """

        # Testing with an 8-block 2x2x2 grid
        grid = GridIO('../data/multi_block/cube/cube.mb.x')
        # Print the doc string
        print(grid)
        grid.read_grid()

        # Simple assertion to test the import
        self.assertEqual(grid.grd.shape, (grid.ni.max(), grid.nj.max(), grid.nk.max(), 3, grid.nb))

    def test_mb_plate(self):
        """
        This tests the padding functionality as well
        Returns:

        """

        # Testing with a multi-block plate grid
        grid = GridIO('../data/multi_block/plate/plate.mb.sp.x')
        # Print the doc string
        print(grid)
        grid.read_grid(data_type='f4')

        # Compute metrics from the test setup function
        test_grid = grid_metrics(grid)

        # Simple assertion to test the import
        self.assertEqual(grid.grd.shape, (grid.ni.max(), grid.nj.max(), grid.nk.max(), 3, grid.nb))
        self.assertEqual(np.all(grid.m1 == test_grid.m1), True)
        self.assertEqual(np.all(grid.m2 == test_grid.m2), True)
        self.assertEqual(np.all(grid.J == test_grid.J), True)

    def test_two_to_three(self):
        """
        Tests the conversion of data from 2d to 3d
        Returns: None

        """

        # Import the shock interaction case
        grid = GridIO('../data/shock_interaction/shock_interaction_coarse.x')
        grid.read_grid()
        grid.two_to_three()

        test_grid = GridIO('../data/shock_interaction/shock_interaction_coarse_3D.x')
        test_grid.read_grid()

    def test_read_formatted_txt(self):
        """
        Tests the code to read formatted text obtained from Tecplot
        Returns:

        """
        path = '../data/shock_interaction/fine/'
        grid = GridIO(path + 'fine_python.x')
        grid.read_grid()

        flow = FlowIO(path + '42500.txt')
        # Fill out required variables for the q-file
        flow.mach = 2.3
        flow.rey = 32033863.98
        flow.alpha = 0.0
        flow.time = 1.0
        flow.read_formatted_txt(grid=grid, data_type='f8')


def grid_metrics(grid):
    """
    Computes grid metrics traditionally
    Returns:
         m1, m2, J
    """

    grid.m1 = np.zeros((grid.ni.max(), grid.nj.max(), grid.nk.max(), 3, 3, grid.nb))
    grid.m2 = np.zeros((grid.ni.max(), grid.nj.max(), grid.nk.max(), 3, 3, grid.nb))
    grid.J = np.zeros((grid.ni.max(), grid.nj.max(), grid.nk.max(), grid.nb))

    # xi derivatives
    for b in range(grid.nb):
        for k in range(grid.nk[b]):
            for j in range(grid.nj[b]):
                grid.m1[0, j, k, 0, 0, b] = grid.grd[1, j, k, 0, b] - grid.grd[0, j, k, 0, b]
                grid.m1[0, j, k, 1, 0, b] = grid.grd[1, j, k, 1, b] - grid.grd[0, j, k, 1, b]
                grid.m1[0, j, k, 2, 0, b] = grid.grd[1, j, k, 2, b] - grid.grd[0, j, k, 2, b]
                for i in range(1, grid.ni[b] - 1):
                    grid.m1[i, j, k, 0, 0, b] = 0.5 * (grid.grd[i + 1, j, k, 0, b] - grid.grd[i - 1, j, k, 0, b])
                    grid.m1[i, j, k, 1, 0, b] = 0.5 * (grid.grd[i + 1, j, k, 1, b] - grid.grd[i - 1, j, k, 1, b])
                    grid.m1[i, j, k, 2, 0, b] = 0.5 * (grid.grd[i + 1, j, k, 2, b] - grid.grd[i - 1, j, k, 2, b])
                grid.m1[grid.ni[b] - 1, j, k, 0, 0, b] = grid.grd[grid.ni[b] - 1, j, k, 0, b] - grid.grd[
                    grid.ni[b] - 2, j, k, 0, b]
                grid.m1[grid.ni[b] - 1, j, k, 1, 0, b] = grid.grd[grid.ni[b] - 1, j, k, 1, b] - grid.grd[
                    grid.ni[b] - 2, j, k, 1, b]
                grid.m1[grid.ni[b] - 1, j, k, 2, 0, b] = grid.grd[grid.ni[b] - 1, j, k, 2, b] - grid.grd[
                    grid.ni[b] - 2, j, k, 2, b]
        print('xi derivatives computed')
        # eta derivatives
        for k in range(grid.nk[b]):
            for i in range(grid.ni[b]):
                grid.m1[i, 0, k, 0, 1, b] = grid.grd[i, 1, k, 0, b] - grid.grd[i, 0, k, 0, b]
                grid.m1[i, 0, k, 1, 1, b] = grid.grd[i, 1, k, 1, b] - grid.grd[i, 0, k, 1, b]
                grid.m1[i, 0, k, 2, 1, b] = grid.grd[i, 1, k, 2, b] - grid.grd[i, 0, k, 2, b]
                for j in range(1, grid.nj[b] - 1):
                    grid.m1[i, j, k, 0, 1, b] = 0.5 * (grid.grd[i, j + 1, k, 0, b] - grid.grd[i, j - 1, k, 0, b])
                    grid.m1[i, j, k, 1, 1, b] = 0.5 * (grid.grd[i, j + 1, k, 1, b] - grid.grd[i, j - 1, k, 1, b])
                    grid.m1[i, j, k, 2, 1, b] = 0.5 * (grid.grd[i, j + 1, k, 2, b] - grid.grd[i, j - 1, k, 2, b])
                grid.m1[i, grid.nj[b] - 1, k, 0, 1, b] = grid.grd[i, grid.nj[b] - 1, k, 0, b] - grid.grd[
                    i, grid.nj[b] - 2, k, 0, b]
                grid.m1[i, grid.nj[b] - 1, k, 1, 1, b] = grid.grd[i, grid.nj[b] - 1, k, 1, b] - grid.grd[
                    i, grid.nj[b] - 2, k, 1, b]
                grid.m1[i, grid.nj[b] - 1, k, 2, 1, b] = grid.grd[i, grid.nj[b] - 1, k, 2, b] - grid.grd[
                    i, grid.nj[b] - 2, k, 2, b]
        print('eta derivatives computed')
        # zeta derivatives
        for j in range(grid.nj[b]):
            for i in range(grid.ni[b]):
                grid.m1[i, j, 0, 0, 2, b] = grid.grd[i, j, 1, 0, b] - grid.grd[i, j, 0, 0, b]
                grid.m1[i, j, 0, 1, 2, b] = grid.grd[i, j, 1, 1, b] - grid.grd[i, j, 0, 1, b]
                grid.m1[i, j, 0, 2, 2, b] = grid.grd[i, j, 1, 2, b] - grid.grd[i, j, 0, 2, b]
                for k in range(1, grid.nk[b] - 1):
                    grid.m1[i, j, k, 0, 2, b] = 0.5 * (grid.grd[i, j, k + 1, 0, b] - grid.grd[i, j, k - 1, 0, b])
                    grid.m1[i, j, k, 1, 2, b] = 0.5 * (grid.grd[i, j, k + 1, 1, b] - grid.grd[i, j, k - 1, 1, b])
                    grid.m1[i, j, k, 2, 2, b] = 0.5 * (grid.grd[i, j, k + 1, 2, b] - grid.grd[i, j, k - 1, 2, b])
                grid.m1[i, j, grid.nk[b] - 1, 0, 2, b] = grid.grd[i, j, grid.nk[b] - 1, 0, b] - grid.grd[
                    i, j, grid.nk[b] - 2, 0, b]
                grid.m1[i, j, grid.nk[b] - 1, 1, 2, b] = grid.grd[i, j, grid.nk[b] - 1, 1, b] - grid.grd[
                    i, j, grid.nk[b] - 2, 1, b]
                grid.m1[i, j, grid.nk[b] - 1, 2, 2, b] = grid.grd[i, j, grid.nk[b] - 1, 2, b] - grid.grd[
                    i, j, grid.nk[b] - 2, 2, b]
        print('zeta derivatives')

        return grid


if __name__ == '__main__':
    unittest.main()



# Calculates different variables from plot3d data
# Equations can be found here: https://www.grc.nasa.gov/WWW/winddocs/towne/plotc/plotc_p3d.html
# TODO: Add more variables
#  Enthalpies, Vorticity, Entropy, Turbulence Parameters, Gradients, Move metrics here?
#  Total quantities
import numpy as np


class Variables:
    """
    Module to compute flow variables

    ...

    Attributes
    ----------
    Input:
        flow : src.io.plot3dio.FlowIO or src.streamlines.interpolation.Interpolation or similar
            object with q --> flow data
        gamma : float
            default is 1.4; can specify
    Output:
        velocity : ndarray
            velocity of the flow at nodes; shape of the flow.q (ni x nj x nk x 3 x nb)
        velocity_magnitude : ndarray
            magnitude of velocity at nodes; shape is (ni x nj x nk x nb)
        temperature : ndarray
            temperature at nodes; shape is (ni x nj x nk x nb)
        pressure : ndarray
            pressure at nodes; shape is (ni x nj x nk x nb)

    Methods
    -------
    compute_velocity()
        computes the velocity and velocity_magnitude
    compute_temperature()
        computes the temperature
    compute_pressure()
        computes the pressure
    compute()
        computes all the variables available

    ...

    Example:
    -------
        variables = Variables(flow)  # Assume flow object is pre-defined
        variables.compute_velocity()  # returns velocity attribute
        variables.compute_temperature()  # fills up the temperature attribute
        variables.compute()  # computes all the attributes available

    """

    def __init__(self, flow, gamma=1.4, gas_constant=287.052874):
        self.flow = flow
        self.gamma = gamma
        self.gas_constant = gas_constant
        self.density = flow.q[..., 0, :]
        self.velocity = None
        self.mach = None
        self.velocity_magnitude = None
        self.temperature = None
        self.pressure = None

    def compute_velocity(self):
        """
        Function to compute velocity and velocity magnitude
        :return: None
        """
        self.velocity = self.flow.q[..., 1:4, :] / (self.flow.q[..., 0, :, None])
        self.velocity_magnitude = (self.velocity[..., 0, :]**2 + self.velocity[..., 1, :]**2 + self.velocity[..., 2, :]**2)**0.5

    def compute_temperature(self):
        """
        Function to compute temperature.
        This computes velocity first
        :return: None
        """
        self.compute_velocity()
        _E_T = self.flow.q[..., 4, :]
        self.temperature = (self.gamma - 1) * (_E_T/self.flow.q[..., 0, :] - self.velocity_magnitude/2) / self.gas_constant

    def compute_mach(self):
        """
        Function to compute local mach number
        Returns: None
        """
        self.compute_temperature()
        self.mach = self.velocity_magnitude / np.sqrt(self.gamma * self.gas_constant * self.temperature)

    def compute_pressure(self):
        """
        Function to compute pressure.
        This computes velocity and temperature first
        :return: None
        """
        self.compute_mach()
        self.pressure = self.flow.q[..., 0, :] * self.temperature * self.gas_constant

    def compute(self):
        # implicitly runs compute_velocity() and compute_temperature()
        """
        Function to compute all the attributes in the class
        :return: None
        """
        self.compute_pressure()



#!/usr/bin/env python3

import numpy as np
import sys

path = '/Users/kal/Library/CloudStorage/OneDrive-UniversityofCincinnati' \
       '/Desktop/University of Cincinnati/DoctoralWork/Codes/project-arrakis'
sys.path.append(path)
import matplotlib.pyplot as plt
from src.streamlines.stochastic_model import StochasticModel, Particle, SpawnLocations
from src.io.plot3dio import GridIO, FlowIO


def test_stochastic_model():
    # Test particle class
    p = Particle()
    p.min_dia = 177e-9
    p.max_dia = 573e-9
    p.mean_dia = 281e-9
    p.std_dia = 97e-9
    p.density = 813
    p.n_concentration = 4
    p.compute_distribution()

    # Test SpawnLocations class
    l = SpawnLocations(p)
    l.x_min = 9e-4
    l.z_min = 2e-4
    l.y_min, l.y_max = 2e-4, 15e-4
    l.compute()

    # Run the model in parallel
    grid_file, flow_file = path + '/data/shocks/shock_test.sb.sp.x', path + '/data/shocks/shock_test.sb.sp.q'
    grid = GridIO(grid_file)
    grid.read_grid()
    grid.compute_metrics()
    flow = FlowIO(flow_file)
    flow.read_flow()
    sm = StochasticModel(p, l, grid=grid, flow=flow)
    sm.method = 'adaptive-ppath'
    sm.drag_model = "henderson"
    sm.search = 'p-space'
    sm.time_step = 1e-10
    sm.max_time_step = 1e-10
    sm.filepath = path + '/data/shocks/particle_data/multi_process_test/'

    # Run multiprocess
    lpt_data = sm.multi_process()


if __name__ == '__main__':
    test_stochastic_model()



import unittest
import numpy as np
import matplotlib.pyplot as plt
from src.streamlines.stochastic_model import StochasticModel, Particle, SpawnLocations
from src.io.plot3dio import GridIO, FlowIO


class TestStochasticModel(unittest.TestCase):
    def test_stochastic_model(self):
        # Test particle class
        p = Particle()
        p.min_dia = 281e-9
        p.max_dia = 281e-9
        p.mean_dia = 281e-9
        p.std_dia = 0
        p.density = 813
        p.n_concentration = 1300
        p.compute_distribution()

        # Test SpawnLocations class
        l = SpawnLocations(p)
        l.x_min = 9e-4
        l.z_min = 2e-4
        l.y_min, l.y_max = 2e-4, 15e-4
        l.compute()

        # Run the model in parallel
        grid_file, flow_file = '../data/shocks/shock_test.sb.sp.x', '../data/shocks/shock_test.sb.sp.q'
        grid = GridIO(grid_file)
        grid.read_grid()
        grid.compute_metrics()
        flow = FlowIO(flow_file)
        flow.read_flow()
        sm = StochasticModel(p, l, grid=grid, flow=flow)
        sm.method = 'adaptive-ppath'
        sm.drag_model = "henderson"
        sm.search = 'p-space'
        sm.time_step = 1e-10
        sm.max_time_step = 1
        sm.adaptivity = 0.001
        # this saves data after every process is done. This will open up memory as well
        sm.filepath = '../data/shocks/particle_data/281nm_time_step_adaptive/'

        # Run multiprocess
        lpt_data = sm.multi_process()

        # save data
        # for i in range(p.n_concentration):
        #     xdata = np.array(lpt_data[i].streamline)
        #     vdata = np.array(lpt_data[i].svelocity)
        #     udata = np.array(lpt_data[i].fvelocity)
        #     data_save = np.hstack((xdata, vdata, udata))
            # np.save('../data/shocks/particle_data/multi_process_test/final_data/' + 'particle_number_' + str(i),
            #         data_save)


if __name__ == '__main__':
    unittest.main()



# Test the data interpolation

import unittest


class TestInterpolation(unittest.TestCase):
    def test_interpolation(self):
        from src.io.plot3dio import GridIO, FlowIO
        from src.streamlines.search import Search
        from src.streamlines.interpolation import Interpolation

        # Read the grid data
        grid = GridIO('../data/plate_data/plate.sp.x')
        flow = FlowIO('../data/plate_data/sol-0000010.q')
        idx = Search(grid, [8.5, 0.5, 0.01])
        point_data = Interpolation(flow, idx)

        grid.read_grid()
        grid.compute_metrics()
        flow.read_flow()
        idx.compute(method='c-space')
        point_data.compute(method='rbf-c-space')

        self.assertEqual(
            sum(abs(point_data.q.reshape(5)
                    - [9.99767442e-01, 1.02352604e-01, -5.38538464e-06, 6.40554753e-09, 1.79096631e+00]))
            <= 1e-6,
            True)


if __name__ == '__main__':
    unittest.main()



# This file contains classes for plot3d data io
# TODO: implement output for p3d data - Assigned to Harpreet.
#  Each output function should be added to respective GridIO/FlowIO classes
#  Change docstrings for doctest in test_io

import numpy as np


class GridIO:
    """Module to read-in a grid file and output grid parameters

    ...

    Attributes
    ----------
    Input :
        filename : str
            name of the grid file
    Output :
        nb : int
            number of blocks in the grid
        ni, nj, nk : int
            shape of the domain
        grd : numpy.ndarray
            Grid data of shape (ni, nj, nk, 3, nb)
        grd_min: numpy.ndarray
            min co-ordinates of each block (3, nb)
        grd_max: numpy.ndarray
            max co-ordinates of each block (3, nb)
        m1 : numpy.ndarray
            xi, eta, zeta derivatives wrt x, y, z --> shape (ni, nj, nk, 3, 3, nb)
        m2 : numpy.ndarray
            x, y, z derivatives wrt xi, eta, zeta --> shape (ni, nj, nk, 3, 3, nb)
        J : numpy.ndarray
            Jacobian determinant of m1 --> shape (ni, nj, nk, nb)


    Methods
    -------
    read_grid()
        returns the output attributes

    Example:
        grid = GridIO('plate.sp.x')  # Assume file is in the path
        grid.read_grid()  # Call method to read the data
        grid.compute_metrics()  # Computes m1, m2, J arrays
        print(grid)  # prints the docstring for grid
        # Instance attributes
        print(grid.grd.shape)  # shape of the grid data
        print(grid.nb)  # Number of blocks

    author: Dilip Kalagotla @ kal ~ dilip.kalagotla@gmail.com
    date: 10-05/2021
    """

    def __init__(self, filename):
        self.filename = filename
        self.nb = None
        self.ni, self.nj, self.nk = None, None, None
        self.grd = None
        self.grd_min, self.grd_max = [], []
        self.m1, self.m2 = None, None
        self.J = None

    def __str__(self):
        doc = "This instance has the filename " + self.filename + "\n" + \
              "grd attribute is of shape (ni, nj, nk, 3, nb), rows representing x,y,z coordinates for each block\n" \
              "For example, x = grid.grd(...,0, 0), grid being the object.\n" \
              "Use method 'read_grid' to compute grd attributes:\n" \
              "ni, nj, nk, ng\n" \
              "Use method 'compute_metrics' to compute grid metric attributes:\n" \
              "m1, m2, J\n" \
              "grd -- The grid data\n"
        return doc

    def read_grid(self, data_type='f4'):
        """Reads in the grid file and changes the instance attributes

        Parameters
        -----------
        data_type: str
            Specify the data type of the grid file specified
            Default is 'f4' for single-precision
            For double-precision use 'f8'

        Returns
        -------
        None

        author: Dilip Kalagotla @ kal ~ dilip.kalagotla@gmail.com
        credit: Paul Orkwis
        date: 11-04/2021
        """
        with open(self.filename, 'r') as grid:
            # Read-in number of blocks
            self.nb = np.fromfile(grid, dtype='i4', count=1)[0]

            # Read-in i, j, k for all blocks
            _temp = np.fromfile(grid, dtype='i4', count=3 * self.nb)
            self.ni, self.nj, self.nk = _temp[0::3], _temp[1::3], _temp[2::3]

            # length of grd data
            _nt = self.ni * self.nj * self.nk * 3
            # read the grd data from file to temp_array
            _temp = np.fromfile(grid, dtype=data_type, count=sum(_nt))

            # pre-define grd to reduce calling pad and concatenate
            self.grd = np.zeros((self.ni.max(), self.nj.max(), self.nk.max(), 3, self.nb))

            # Reshape and assign data to grd
            for _i in range(self.nb):
                self.grd[0:self.ni[_i], 0:self.nj[_i], 0:self.nk[_i], 0:3, _i] = \
                    _temp[sum(_nt[0:_i]):sum(_nt[0:_i]) + _nt[_i]] \
                    .reshape((self.ni[_i], self.nj[_i], self.nk[_i], 3), order='F')

            print("Grid data reading is successful for " + self.filename + "\n")

            # Setup some parameters for further processing
            # Find out the min and max of each block
            for _i, _j, _k, _b in zip(self.ni, self.nj, self.nk, range(self.nb)):
                self.grd_min.append(np.amin(self.grd[:_i, :_j, :_k, :, _b], axis=(0, 1, 2)))
                self.grd_max.append(np.amax(self.grd[:_i, :_j, :_k, :, _b], axis=(0, 1, 2)))

            # Convert lists to arrays
            self.grd_min = np.array(self.grd_min)
            self.grd_max = np.array(self.grd_max)

    def compute_metrics(self):
        """Calculate grid metrics from grid data.
        Need to call read_grid() before computing metrics

        Parameters
        ----------

        Returns
        -------
        None

        Example:
            grid = GridIO(filename)  # Assume grid is the object from GridIO
            grid.read_grid()  # Call method to read-in grid data
            grid.compute_metrics()  # Call method to compute grid metrics
            print(grid)  # prints the docstring for grid metrics
            # Instance attributes
            print(grid.m1)  # derivatives xi, eta, zeta
            print(grid.m2)  # derivatives x, y, z

        author: Dilip Kalagotla @ kal ~ dilip.kalagotla@gmail.com
        credit: Jacob Welsh for providing the code for single block
        date: 12-23/2021
        """
        self.m1 = np.zeros((self.ni.max(), self.nj.max(), self.nk.max(), 3, 3, self.nb))
        self.m2 = np.zeros((self.ni.max(), self.nj.max(), self.nk.max(), 3, 3, self.nb))
        self.J = np.zeros((self.ni.max(), self.nj.max(), self.nk.max(), self.nb))

        #  The for loop is to compute x, y, z derivatives wrt xi, eta, zeta
        # for i in range(3):
        #     self.m1[..., i, :] = np.gradient(self.grd, axis=i)

        for b in range(self.nb):
            print(f"Computing Jacobian for block {b}...")
            for i in range(3):
                self.m1[:self.ni[b], :self.nj[b], :self.nk[b], :, i, b] = \
                    np.gradient(self.grd[:self.ni[b], :self.nj[b], :self.nk[b], :, b], axis=i)
        print("Done computing Jacobian for all the blocks!!\n")

        # compute Jacobian
        for b in range(self.nb):
            print(f"Computing Jacobian determinant for block {b}...")
            self.J[:self.ni[b], :self.nj[b], :self.nk[b], b] = \
                self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 0, b] * \
                (self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 1, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 2, b] -
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 2, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 1, b]) - \
                self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 0, b] * \
                (self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 1, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 2, b] -
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 2, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 1, b]) + \
                self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 0, b] * \
                (self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 1, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 2, b] -
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 2, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 1, b])
        print("Done computing Jacobian determinant for all the blocks!!")

        # x derivatives
        for b in range(self.nb):
            print(f"Computing Inverse Jacobian for block {b}...")
            self.m2[:self.ni[b], :self.nj[b], :self.nk[b], 0, 0, b] = \
                (self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 1, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 2, b] -
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 2, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 1, b]) /\
                self.J[:self.ni[b], :self.nj[b], :self.nk[b], b]
            self.m2[:self.ni[b], :self.nj[b], :self.nk[b], 0, 1, b] = \
                (self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 2, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 1, b] -
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 1, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 2, b]) /\
                self.J[:self.ni[b], :self.nj[b], :self.nk[b], b]
            self.m2[:self.ni[b], :self.nj[b], :self.nk[b], 0, 2, b] = \
                (self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 1, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 2, b] -
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 2, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 1, b]) /\
                self.J[:self.ni[b], :self.nj[b], :self.nk[b], b]

            # y derivative
            self.m2[:self.ni[b], :self.nj[b], :self.nk[b], 1, 0, b] = \
                (self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 2, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 0, b] -
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 0, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 2, b]) /\
                self.J[:self.ni[b], :self.nj[b], :self.nk[b], b]
            self.m2[:self.ni[b], :self.nj[b], :self.nk[b], 1, 1, b] = \
                (self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 0, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 2, b] -
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 2, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 0, b]) /\
                self.J[:self.ni[b], :self.nj[b], :self.nk[b], b]
            self.m2[:self.ni[b], :self.nj[b], :self.nk[b], 1, 2, b] = \
                (self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 2, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 0, b] -
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 0, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 2, b]) /\
                self.J[:self.ni[b], :self.nj[b], :self.nk[b], b]

            # z derivatives
            self.m2[:self.ni[b], :self.nj[b], :self.nk[b], 2, 0, b] = \
                (self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 0, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 1, b] -
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 1, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 0, b]) /\
                self.J[:self.ni[b], :self.nj[b], :self.nk[b], b]
            self.m2[:self.ni[b], :self.nj[b], :self.nk[b], 2, 1, b] = \
                (self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 1, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 0, b] -
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 0, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 2, 1, b]) /\
                self.J[:self.ni[b], :self.nj[b], :self.nk[b], b]
            self.m2[:self.ni[b], :self.nj[b], :self.nk[b], 2, 2, b] = \
                (self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 0, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 1, b] -
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 0, 1, b] *
                 self.m1[:self.ni[b], :self.nj[b], :self.nk[b], 1, 0, b]) /\
                self.J[:self.ni[b], :self.nj[b], :self.nk[b], b]

        print("Done computing Inverse Jacobian for all the blocks!!")
        print("All Grid metrics computed successfully!")

    def two_to_three(self, steps: int = 5, step_size: float = None, data_type='f4'):
        """
        Converts 2D plot3d grid file to 3D format
        TODO: Current limitation is that the code works only for 3d written 2d files and single block
        Returns: None

        """
        if step_size is None:
            print("Step size is not provided; Using minimum grid size")
            step_size = abs(min(np.diff(self.grd[:, 0, 0, 0, 0])))

        _a_temp = np.array([self.nb, self.ni, self.nj, steps], dtype='i4')
        _x_temp = self.grd[..., 0, 0].repeat(steps, axis=2)
        _y_temp = self.grd[..., 1, 0].repeat(steps, axis=2)
        _z_temp = np.ones((int(self.ni), int(self.nj), steps)) * np.linspace(0, steps*step_size, steps)
        _b_temp = np.array([_x_temp.T, _y_temp.T, _z_temp.T], dtype=data_type)

        _temp_filename = self.filename.replace('.x', '_3D.x')
        with open(_temp_filename, 'wb') as f:
            f.write(_a_temp.tobytes())
            f.write(_b_temp.tobytes())

        print(f'\n File is successfully written in the working directory as {_temp_filename}')

        return

    def mgrd_to_p3d(self, xi, yi, out_file: str = 'mgrd_to_p3d.x',
                    steps: int = 5, step_size: float = None, data_type='f4'):
        """
        Creates a 3d plot3d grid file;
        This can be later used to be converted to 3d format by two_to_three
        Returns:

        """
        if step_size is None:
            print("Step size is not provided; Using minimum grid size")
            step_size = abs(min(np.diff(self.grd[:, 0, 0, 0, 0])))

        # Number of blocks is always 1 for this function
        _ng, _ni, _nj, _nk = np.array([1, xi.shape[0], yi.shape[1], steps], dtype='i4')
        _xx, _yy, _zz = np.meshgrid(xi[:, 0], yi[0], np.linspace(0, steps*step_size, steps), indexing='ij')
        _grd = np.array([_xx.T, _yy.T, _zz.T], dtype=data_type)

        with open(out_file, 'wb') as f:
            f.write(_ng)
            f.write(_ni)
            f.write(_nj)
            f.write(_nk)
            f.write(_grd.tobytes())

        print(f'\n File is successfully written in the working directory as {out_file}')

        return


class FlowIO:
    """Module to read-in a flow file and output flow parameters

    ...

    Attributes
    ----------
    Input :
        filename : str
            name of the flow file
    Output :
        nb : int
            number of blocks in the grid
        ni, nj, nk : int
            shape of the domain
        mach, alpha, rey, time: float
            extra numerics needed for post-processing
        q : numpy.ndarray
            Flow data of shape (ni, nj, nk, 5, nb)

    Methods
    -------
    read_flow()
        returns the output attributes
    two_to_three()
        converts 2D data to 3D
    mgrd_to_p3d()
        converts and returns meshgrid data to plot3D data
    read_formatted_txt()
        Reads Tecplot returned formatted text. Must contain five columns without headers
        rho, rho-u, rho-v, rho-w, e are the five columns

    Example:
        grid = FlowIO('plate.sp.x')  # Assume file is in the path\n
        flow.read_flow()  # Call method to read the data\n
        print(flow)  # prints the docstring for grid\n
        # Instance attributes\n
        print(flow.q.shape)  # shape of the flow data\n
        print(flow.ng)  # Number of blocks


    author: Dilip Kalagotla @ kal ~ dilip.kalagotla@gmail.com
    date: 10-05/2021
    """

    def __init__(self, filename):
        self.filename = filename
        self.nb = None
        self.ni, self.nj, self.nk = None, None, None
        self.mach = None
        self.alpha = None
        self.rey = None
        self.time = None
        self.q = None

    def __str__(self):
        doc = "This instance has the filename " + self.filename + "\n" + \
              "q attribute is of shape (ni, nj, nk, 5, nb), rows representing density, momentum[*3], energy" \
              "for every block\n" \
              "For example, rho = flow.q(...,0,0), flow being the object for block-1.\n" \
              "Use method 'read_flow' to compute attributes:\n" \
              "nb, ni, nj, nk\n" \
              "mach, alpha, rey, time\n" \
              "q -- The flow data\n"
        return doc

    def read_flow(self, data_type='f4'):
        """Reads in the flow file and changes the instance attributes

        Parameters
        ----------
        data_type: str
            Specify the data type of the flow file specified
            Default is 'f4' for single-precision
            For double-precision use 'f8'

        Returns
        -------
        None

        author: Dilip Kalagotla @ kal ~ dilip.kalagotla@gmail.com
        credit: Paul Orkwis
        date: 10-05/2021
        """
        with open(self.filename, 'r') as data:
            self.nb = np.fromfile(data, dtype='i4', count=1)[0]

            # Read in the i, j, k values for blocks
            _temp = np.fromfile(data, dtype='i4', count=3 * self.nb)
            self.ni, self.nj, self.nk = _temp[0::3], _temp[1::3], _temp[2::3]

            # Read-in flow data into a temp array
            # Less use of fromfile is more speed
            _nt = self.ni * self.nj * self.nk * 5
            _temp = np.fromfile(data, dtype=data_type, count=sum(_nt) + 4 * self.nb)

            # Assign the dimensionless attributes
            self.mach, self.alpha, self.rey, self.time = _temp[0:4]

            # Create a mask to remove dimensionless quantities from q
            # Indices of the first four dimensionless quantities
            _index_array = np.array([0, 1, 2, 3])
            for i in range(len(_nt) - 1):
                _term = sum(_temp[0:i]) + (i + 1) * 4
                _index_array = np.concatenate((_index_array, np.arange(_term, _term + 4)))
            _mask = np.ones(len(_temp), dtype='bool')
            _mask[_index_array] = 0
            # Use the mask to remove dimensionless quantities
            _temp = _temp[_mask]

            # Pre-define q array
            self.q = np.zeros((self.ni.max(), self.nj.max(), self.nk.max(), 5, self.nb))

            # Reshape and assign data to q
            for _i in range(self.nb):
                self.q[0:self.ni[_i], 0:self.nj[_i], 0:self.nk[_i], 0:5, _i] = \
                    _temp[sum(_nt[0:_i]):sum(_nt[0:_i]) + _nt[_i]] \
                    .reshape((self.ni[_i], self.nj[_i], self.nk[_i], 5), order='F')

            print("Flow data reading is successful for " + self.filename + "\n")

    def two_to_three(self, steps: int = 5, data_type='f4'):
        """
        Converts 2D plot3d flow file to 3D format
        TODO: Current limitation is that the code works only for 3d written 2d files and single block
        Returns: None

        """
        # TODO: The code below needs debugging. It's not tested. mgrd_to_p3d might help!
        _a_temp = np.array([self.nb, self.ni, self.nj, int(steps)], dtype='i4')
        _b_temp = np.array([self.mach, self.alpha, self.rey, self.time], dtype=data_type)
        _q0_temp = self.q[..., 0, 0].repeat(int(steps), axis=2)
        _q1_temp = self.q[..., 1, 0].repeat(int(steps), axis=2)
        _q2_temp = self.q[..., 2, 0].repeat(int(steps), axis=2)
        _q3_temp = self.q[..., 3, 0].repeat(int(steps), axis=2)
        _q4_temp = self.q[..., 4, 0].repeat(int(steps), axis=2)
        _q_temp = np.array([_q0_temp, _q1_temp, _q2_temp, _q3_temp, _q4_temp], dtype=data_type)

        _temp_filename = self.filename.replace('.q', '_3D.q')
        with open(_temp_filename, 'wb') as f:
            f.write(_a_temp.tobytes())
            f.write(_b_temp.tobytes())
            f.write(_q_temp.tobytes())

        return

    def mgrd_to_p3d(self, q, out_file: str = 'mgrd_to_p3d', mode: str = None, steps: int = 5, data_type='f4'):
        """
        Writes out data generated from scattered data interpolation to plot3d format
        Args:
            q:
            out_file:
            mode:
            steps:
            data_type:

        Returns:

        """
        _a_temp = np.array([1, q.shape[1], q.shape[-1], int(steps)], dtype='i4')
        _b_temp = np.array([self.mach, self.alpha, self.rey, self.time], dtype=data_type)
        _q0 = np.expand_dims(q[0, ...], axis=2).T.repeat(int(steps), axis=0)
        _q1 = np.expand_dims(q[1, ...], axis=2).T.repeat(int(steps), axis=0)
        _q2 = np.expand_dims(q[2, ...], axis=2).T.repeat(int(steps), axis=0)
        _q3 = np.expand_dims(q[3, ...], axis=2).T.repeat(int(steps), axis=0)
        _q4 = np.expand_dims(q[4, ...], axis=2).T.repeat(int(steps), axis=0)
        _q = np.array([_q0, _q1, _q2, _q3, _q4], dtype=data_type)

        with open(out_file+'_'+mode+'.q', 'wb') as f:
            f.write(_a_temp.tobytes())
            f.write(_b_temp.tobytes())
            f.write(_q.tobytes())

        return

    def read_formatted_txt(self, grid, data_type='f8'):
        import os
        """
        Reads the formatted flow file generated from Tecplot. Needs grid object
        Generate data from tecplot without the header, delimiter as space and have 5 variables
        rho, rho-u, rho-v, rho-w, and e
        Manually add all the required variables for flow object
        Args:
            grid: grid object related to the flow file
            data_type: data type of the formatted text from Tecplot

        Returns:
            None

        """
        try:
            # load the flow file if available
            self.q = np.load(self.filename + '_temp/flow_data.npy')
            print('**IMPORTANT** Read data from existing temp flow_data.npy file.\n')
        except:
            # Read the formatted data till the file ends and reshape it to (ni, nj, nk, 5, nb)
            print('Starting flow read from the formatted text. Please wait...\n')
            with open(self.filename, 'r') as flow:
                self.q = np.fromfile(flow, sep=' ', dtype=data_type, count=-1)\
                    .reshape((int(grid.nk), int(grid.nj), int(grid.ni), 5, 1)).transpose(2, 1, 0, 3, 4)

            # Save as numpy file for future computations
            try:
                # Try creating the directory; if exists errors out and except
                print('Trying to save to npy for future use\n')
                os.mkdir(self.filename + '_temp')
                np.save(self.filename + '_temp/flow_data', self.q)
                print('Created _temp folder and saved flow data for future use.\n')
            except:
                np.save(self.filename + '_temp/flow_data', self.q)
                print('Saved file for future use.\n')

        # Fill in other variables
        if self.mach is not None and self.rey is not None and self.alpha is not None and self.time is not None:
            print('\nYour flow object is ready for further computations!\n')
        else:
            print('\nPlease fill out mach, rey, alpha, and time variables in the object\n')
            print('\n** ERROR **: Try again; one of the required variables is not filled\n')
            exit()

        return




import unittest
import numpy as np
import matplotlib.pyplot as plt


class TestPlot(unittest.TestCase):
    def test_plot(self):
        # filepath = '../data/shocks/particle_data/random_seed_7/raw_data/'
        # filepath = '../data/shocks/particle_data/random_seed_7_time_step_adaptive/'
        filepath = '../data/shock_interaction/final_grid/particle_data/'
        fig, ax = plt.subplots(2, 2)
        fig1, ax1 = plt.subplots(2, 2)
        fig2 = plt.figure()
        ax2 = plt.axes(projection='3d')
        for i in range(5):
            # data = np.load(filepath + 'particle_number_' + str(i) + '.npy')
            data = np.load(filepath + 'ppath_' + str(i) + '.npy')
            xp, yp, zp = data[:, 0], data[:, 1], data[:, 2]
            vx, vy, vz = data[:, 3], data[:, 4], data[:, 5]
            ux, uy, uz = data[:, 6], data[:, 7], data[:, 8]
            time = data[:, 9]
            xf, yf, zf = data[:, 10], data[:, 11], data[:, 12]

            # Plot in a single window
            ax[0, 0].plot(xp, yp, 'r')
            ax[0, 0].plot(xf, yf, 'b')
            ax[0, 0].set_title('paths')
            ax[0, 1].plot(xp, vx, '.-r')
            ax[0, 1].plot(xp, ux, '.-b')
            ax[0, 1].set_title('x-velocity')
            ax[1, 0].plot(xp, vy, 'r')
            ax[1, 0].plot(xp, uy, 'b')
            ax[1, 0].set_title('y-velocity')
            ax[1, 1].plot(xp, vz, 'r')
            ax[1, 1].plot(xp, uz, 'b')
            ax[1, 1].set_title('z-velocity')

            print('Error in particle to fluid path deviation: ', np.linalg.norm(data[:, 0:3] - data[:, 3:6]))

            ax1[0, 0].plot(xp, abs(xp-xf), 'k')
            ax1[0, 0].set_title('Difference in paths')
            ax1[0, 1].plot(xp, abs(vx-ux), 'k')
            ax1[0, 1].set_title('Difference in x-velocity')

            ax2.plot3D(xp, yp, zp, 'r')
            ax2.plot3D(xf, yf, zf, 'b')
            ax2.set_xlim(0, 0.005)
            ax2.set_ylim(0, 0.003)
            ax2.set_zlim(0, 0.001)
        plt.show()


if __name__ == '__main__':
    unittest.main()



# Uses the program API to extract streamlines
import numpy as np
from src.function.timer import Timer
from src.io.plot3dio import GridIO
from src.io.plot3dio import FlowIO
from src.streamlines.search import Search
from src.streamlines.interpolation import Interpolation
from src.streamlines.integration import Integration
from src.function.variables import Variables


class Streamlines:
    """
    Module to extract particle path information

    ...

    Attributes
    ----------
    Input:
        grid_file : str
            Path to the plot3d grid data file
        flow_file : str
            Path to the plot3d flow data file
        point : list
            Starting point for integration
        search : str
            Default is p-space; can specify c-space
        interpolation : str
            Default is p-space; can specify c-space
        integration : str
            Default is p-space; can specify c-space
        time_step : float
            Default is 1e-3
        magnitude_adaptivity : float
            Set to override the adaptivity value

    Output:
        streamline : numpy.ndarray
            shape is nx3; each column represents x, y, z
        fvelocity : list
            shape is nx3; each column represents ufx, ufy, ufz
        svelocity: list
            shape is nx3; each column represents vpx, vpy, vpz

    Methods
    -------
    compute()
        integrates and returns the streamline ndarray
    ...

    Example:
        sl = Streamlines('../../data/vortex/vortex.sb.sp.x', '../../data/vortex/vortex.sb.sp.q', [-0.05, 0.05, 5])
        sl.compute()
    -------


    """

    def __init__(self, grid_file=None, flow_file=None, point=None,
                 search='p-space', interpolation='p-space', integration='pRK4',
                 diameter=1e-7, density=1000,
                 time_step=1e-3, max_time_step=1, drag_model='stokes', adaptivity=0.001,
                 magnitude_adaptivity=None, filepath=None, task=None):
        self.grid_file = grid_file
        self.flow_file = flow_file
        self.point = np.array(point)
        self.search = search
        self.interpolation = interpolation
        self.integration = integration
        self.diameter = diameter
        self.density = density
        self.time_step = time_step
        self.max_time_step = max_time_step
        self.drag_model = drag_model
        self.streamline = []
        self.fvelocity = []
        self.svelocity = []
        self.time = []
        self.adaptivity = adaptivity
        self.magnitude_adaptivity = magnitude_adaptivity
        self.filepath = filepath
        self.task = task

    # TODO: Need to add doc for streamlines

    @staticmethod
    def angle_btw(v1, v2):
        # If vectors are tiny return for handling tiny time steps
        # A lot of trail and error decided 1e-9 would work fast
        if np.linalg.norm(v1) <= 1e-9 or np.linalg.norm(v2) <= 1e-9:
            return None

        u1 = v1 / np.linalg.norm(v1)
        u2 = v2 / np.linalg.norm(v2)

        y = u1 - u2
        x = u1 + u2

        a0 = 2 * np.arctan(np.linalg.norm(y) / np.linalg.norm(x))

        if (not np.signbit(a0)) or np.signbit(np.pi - a0):
            return np.rad2deg(a0)
        elif np.signbit(a0):
            return 0.0
        else:
            return 180

    @staticmethod
    def _save_data(self):
        """
        Pass self object to save data if the filepath is provided
        Args:
            self: self object of the class

        Returns:
            None
        """
        if self.filepath is not None:
            p_xdata = np.array(self.streamline)
            vdata = np.array(self.svelocity)
            udata = np.array(self.fvelocity)
            tdata = np.array(self.time).reshape(-1, 1)
            f_xdata = p_xdata.copy()
            for _i in range(1, len(f_xdata)):
                f_xdata[_i] = f_xdata[_i - 1] + udata[_i - 1] * tdata[_i - 1]
            # Data is added towards the end because of the development cycle. Mostly to work with dataio
            _data_save = np.hstack((p_xdata, vdata, udata, tdata, f_xdata))

            np.save(self.filepath + 'ppath_' + str(self.task), _data_save)
            print('** SUCCESS ** Done writing file for particle number - ' + str(self.task) + ' ** SUCCESS **')
            # set self to None to clear up memory after saving required data
            self.streamline = []
            self.svelocity = []
            self.fvelocity = []
            self.time = []
        return

    @staticmethod
    def _magnitude(self, v1, v2):
        # return the magnitude of the vector
        _r = np.linalg.norm(v1 - v2) / min(np.linalg.norm(v1), np.linalg.norm(v2))
        if _r < 1e-6 * self.magnitude_adaptivity:
            _r = 1e-6 * self.magnitude_adaptivity
        return _r

    def compute(self, method='p-space', grid=None, flow=None):
        """
        Method to compute particle paths. Contains multiple algorithms
        Args:
            method:
            grid:
            flow:

        Returns:

        """
        if grid is None or flow is None:
            grid = GridIO(self.grid_file)
            flow = FlowIO(self.flow_file)

            # Read in the grid and flow data
            grid.read_grid()
            flow.read_flow()
            grid.compute_metrics()

        if self.magnitude_adaptivity is None:
            self.magnitude_adaptivity = abs(np.min(np.gradient(grid.grd[..., 0])))
        # Add data to output at the given point
        # This is the assumption where particle velocity is same as the fluid
        self.streamline.append(self.point)
        idx = Search(grid, self.point)
        interp = Interpolation(flow, idx)
        idx.compute(method=self.search)
        interp.compute(method=self.interpolation)
        q_interp = Variables(interp)
        q_interp.compute_velocity()
        uf = q_interp.velocity.reshape(3)
        self.fvelocity.append(uf)
        self.svelocity.append(uf)
        self.time.append(self.time_step)
        loop_check = 0

        if method == 'p-space':
            t = Timer(text="Time taken for particle " + str(self.task) + " is {:.2f} seconds")
            t.start()
            while True:
                idx = Search(grid, self.point)
                interp = Interpolation(flow, idx)
                intg = Integration(interp)
                idx.compute(method=self.search)
                interp.compute(method=self.interpolation)
                new_point, new_vel = intg.compute(method=self.integration, time_step=self.time_step)
                if new_point is None:
                    print('Integration complete!')
                    break
                self.streamline.append(new_point)
                self.fvelocity.append(new_vel)
                self.svelocity.append(new_vel)
                self.time.append(self.time_step)
                self.point = new_point

            # Save files for each particle; can be used for multiprocessing large number of particles
            self._save_data(self)
            t.stop()

        if method == 'adaptive-p-space':
            t = Timer(text="Time taken for particle " + str(self.task) + " is {:.2f} seconds")
            t.start()
            vel = None
            while True:
                idx = Search(grid, self.point)
                interp = Interpolation(flow, idx)
                intg = Integration(interp)
                idx.compute(method=self.search)
                interp.compute(method=self.interpolation)
                new_point, new_vel = intg.compute(method=self.integration, time_step=self.time_step)
                if new_point is None:
                    print('Checking if the end of the domain is reached...')
                    self.time_step = 1e-9 * self.time_step
                    new_point, new_vel = intg.compute(method=self.integration, time_step=self.time_step)
                    if new_point is not None:
                        print('Continuing integration by decreasing time-step!')
                        continue
                    elif new_point is None:
                        print('Integration complete!')
                        break

                # Adaptive algorithm starts
                # Save results and adjust time-step
                # Details for the algorithm are provided in adaptive-ppath
                if vel is None:
                    # For the first step in the loop
                    self.streamline.append(new_point)
                    self.fvelocity.append(new_vel)
                    self.svelocity.append(new_vel)
                    self.time.append(self.time_step)
                    self.point = new_point
                    vel = new_vel.copy()
                elif self.angle_btw(new_point - self.point, vel) is None:
                    print('Increasing time step. Successive points are same')
                    self.time_step = 10 * self.time_step
                    loop_check += 1
                    if loop_check == 70:
                        print('Stuck in the same loop for too long. Integration ends!')
                        break
                elif self.angle_btw(new_point - self.point, vel) <= 0.001 \
                        and self.time_step <= self.max_time_step:
                    self.streamline.append(new_point)
                    self.fvelocity.append(new_vel)
                    self.svelocity.append(new_vel)
                    self.time.append(self.time_step)
                    self.point = new_point
                    vel = new_vel.copy()
                    self.time_step = 2 * self.time_step
                    loop_check = 0
                elif self.angle_btw(new_point - self.point, vel) >= 0.01 and self.time_step >= 1e-12:
                    self.time_step = 0.5 * self.time_step
                else:
                    self.streamline.append(new_point)
                    self.fvelocity.append(new_vel)
                    self.svelocity.append(new_vel)
                    self.time.append(self.time_step)
                    self.point = new_point
                    vel = new_vel.copy()
                    loop_check = 0

            # Save files for each particle; can be used for multiprocessing large number of particles
            self._save_data(self)
            t.stop()

        if method == 'c-space':
            # Use c-space search to convert and find the location of given point
            # All the idx attributes are converted to c-space -- point, cell, block
            t = Timer(text="Time taken for particle " + str(self.task) + " is {:.2f} seconds")
            t.start()
            save_point = self.point
            idx = Search(grid, self.point)
            idx.compute(method='c-space')
            while True:
                interp = Interpolation(flow, idx)
                interp.compute(method='c-space')
                intg = Integration(interp)
                new_point, new_pvel, new_cvel = intg.compute(method='cRK4', time_step=self.time_step)
                if new_point is None:
                    # For multi-block case if the point is out-of-block
                    # Use previous point and run one-step of p-space algo
                    print('Point exited the block! Searching for new position...')
                    idx = Search(grid, save_point)
                    interp = Interpolation(flow, idx)
                    intg = Integration(interp)
                    idx.compute(method='block_distance')
                    interp.compute()
                    new_point, new_pvel = intg.compute(method='pRK4', time_step=self.time_step)
                    if new_point is None:
                        print('Point out-of-domain. Integration complete!')
                        break
                    else:
                        # Update the block in idx
                        # new_point is in p-space for this else block

                        idx = Search(grid, new_point)
                        idx.compute(method='c-space')
                        self.streamline.append(new_point)
                        self.fvelocity.append(new_pvel)
                        self.svelocity.append(new_pvel)
                        self.time.append(self.time_step)
                        self.point = save_point
                        save_point = new_point
                        pvel = new_pvel.copy()
                else:
                    save_point = idx.c2p(new_point)
                    self.streamline.append(save_point)
                    self.fvelocity.append(new_pvel)
                    self.svelocity.append(new_pvel)
                    self.time.append(self.time_step)
                    idx.point = new_point

            # Save files for each particle; can be used for multiprocessing large number of particles
            self._save_data(self)
            t.stop()

        if method == 'adaptive-c-space':
            # Use c-space search to convert and find the location of given point
            # All the idx attributes are converted to c-space -- point, cell, block
            # pvel signifies physical-space velocity
            t = Timer(text="Time taken for particle " + str(self.task) + " is {:.2f} seconds")
            t.start()
            save_point = self.point
            idx = Search(grid, self.point)
            idx.compute(method='c-space')
            pvel = None
            while True:
                interp = Interpolation(flow, idx)
                interp.compute(method='c-space')
                intg = Integration(interp)
                new_point, new_pvel, new_cvel = intg.compute(method='cRK4', time_step=self.time_step)

                # Check for large time-step in the last loop
                if new_point is None:
                    # Checking by decreasing time-step
                    print('Checking if the end of the domain is reached...')
                    self.time_step = 1e-9 * self.time_step
                    new_point, new_pvel, new_cvel = intg.compute(method='cRK4', time_step=self.time_step)
                    if new_point is not None:
                        print('Continuing integration by decreasing time-step!')
                        continue
                    # Check for block switch
                    elif new_point is None:
                        # For multi-block case if the point is out-of-block
                        # Use previous point and run one-step of p-space algo
                        print('Point exited the block! Searching for new position...')
                        idx = Search(grid, save_point)
                        interp = Interpolation(flow, idx)
                        intg = Integration(interp)
                        idx.compute(method='block_distance')
                        interp.compute()
                        new_point, new_pvel = intg.compute(method='pRK4', time_step=self.time_step)
                        # Even after pRK4 if the point is None; End integration
                        if new_point is None:
                            print('End of the domain. Integration Ends!')
                            break
                        # If not none; update to c-space and run
                        else:
                            # Update the block in idx
                            idx = Search(grid, new_point)
                            idx.compute(method='c-space')
                            self.streamline.append(new_point)
                            self.fvelocity.append(new_pvel)
                            self.svelocity.append(new_pvel)
                            self.time.append(self.time_step)
                            self.point = save_point
                            save_point = new_point
                            pvel = new_pvel.copy()
                            # new_point = idx.p2c(new_point)  # Move point obtained to c-space
                # If the point is not none; continue c-space
                else:
                    self.point = save_point
                    save_point = idx.c2p(new_point)

                    # Adaptive algorithm starts
                    # Save results and adjust time-step
                    # Details for the algorithm are provided in adaptive-ppath
                    # Decided to use new_pvel for adaptivity because of the in-shock cell RK4 integration
                    # RK4 depends on future points for integration and this adjustment moves it forward
                    if self.angle_btw(save_point - self.point, new_pvel) is None:
                        print('Increasing time step. Successive points are same')
                        self.time_step = 2 * self.time_step
                        loop_check += 1
                        if loop_check == 70:
                            print('Stuck in the same loop for too long. Integration ends!')
                            break
                    elif self.angle_btw(save_point - self.point, new_pvel) <= 0.1 * self.adaptivity \
                            and self.time_step <= self.max_time_step:
                        self.point = save_point
                        save_point = idx.c2p(new_point)
                        self.streamline.append(save_point)
                        self.fvelocity.append(new_pvel)
                        self.svelocity.append(new_pvel)
                        self.time.append(self.time_step)
                        idx.point = new_point
                        pvel = new_pvel.copy()
                        self.time_step = 2 * self.time_step
                        loop_check = 0
                    elif self.angle_btw(save_point - self.point, new_pvel) >= self.adaptivity and\
                            self.time_step >= 1e-12:
                        self.time_step = 0.5 * self.time_step
                    else:
                        self.point = save_point
                        save_point = idx.c2p(new_point)
                        self.streamline.append(save_point)
                        self.fvelocity.append(new_pvel)
                        self.svelocity.append(new_pvel)
                        self.time.append(self.time_step)
                        idx.point = new_point
                        pvel = new_pvel.copy()
                        loop_check = 0
            # Save files for each particle; can be used for multiprocessing large number of particles
            self._save_data(self)
            t.stop()

        if method == 'ppath':
            t = Timer(text="Time taken for particle " + str(self.task) + " is {:.2f} seconds")
            t.start()
            vel = None
            fvel = None
            while True:
                idx = Search(grid, self.point)
                interp = Interpolation(flow, idx)
                intg = Integration(interp)
                idx.compute(method=self.search)
                interp.compute(method=self.interpolation)
                new_point, new_vel, new_fvel = intg.compute_ppath(diameter=self.diameter,
                                                                  density=self.density,
                                                                  velocity=vel, method='pRK4',
                                                                  time_step=self.time_step,
                                                                  drag_model=self.drag_model)
                if new_point is None:
                    print('Integration complete!')
                    break

                # Save results and continue the loop
                self.streamline.append(new_point)
                self.svelocity.append(new_vel)
                self.fvelocity.append(new_fvel)
                self.time.append(self.time_step)
                self.point = new_point
                vel = new_vel.copy()
                fvel = new_fvel.copy()

            # Save files for each particle; can be used for multiprocessing large number of particles
            self._save_data(self)
            t.stop()

        if method == 'adaptive-ppath':
            t = Timer(text="Time taken for particle " + str(self.task) + " is {:.2f} seconds")
            t.start()
            # particle velocity
            vel = None
            # fluid velocity
            fvel = None
            while True:
                idx = Search(grid, self.point)
                interp = Interpolation(flow, idx)
                intg = Integration(interp)
                idx.compute(method=self.search)
                interp.compute(method=self.interpolation)
                new_point, new_vel, new_fvel = intg.compute_ppath(diameter=self.diameter, density=self.density,
                                                                  velocity=vel,
                                                                  method='pRK4', time_step=self.time_step,
                                                                  drag_model=self.drag_model)
                if new_point is None:
                    # print('Checking if the end of the domain is reached...')
                    self.time_step = 1e-9 * self.time_step
                    new_point, new_vel, new_fvel = intg.compute_ppath(diameter=self.diameter, density=self.density,
                                                                      velocity=vel,
                                                                      method='pRK4', time_step=self.time_step,
                                                                      drag_model=self.drag_model)
                    if new_point is not None:
                        # print('Continuing integration by decreasing time-step!')
                        continue
                    elif new_point is None:
                        print('Integration complete!')
                        break

                # Check for mid-rk4 blowup
                if intg.rk4_bool is True:
                    print(f'**WARNING** Large residual. Mid-RK4 blow up! Reducing time-step for particle number'
                          f' - {self.task}')
                    intg.rk4_bool = False
                    self.time_step = 0.5 * self.time_step
                    loop_check += 1
                    if loop_check == 70:
                        print('Stuck in the same loop for too long. Integration ends!')
                        break

                # Adaptive algorithm starts
                # Save results and continue the loop
                elif vel is None:
                    # For the first step in the loop
                    self.streamline.append(new_point)
                    self.svelocity.append(new_vel)
                    self.fvelocity.append(new_fvel)
                    self.time.append(self.time_step)
                    self.point = new_point
                    vel = new_vel.copy()
                    fvel = new_fvel.copy()
                # Check for if the points are identical because of tiny time step and deflection
                elif self.angle_btw(new_vel, vel) is None:
                    # print('Increasing time step. Successive points are same')
                    self.time_step = 2 * self.time_step
                    loop_check += 1
                    if loop_check == 70:
                        print(f'Successive points did not change for too long. Integration ends! for particle '
                              f'{self.task}')
                        break
                # Check for strong acceleration and reduce time-step
                # Increase time step when angle is below 0.05 degrees
                elif self.angle_btw(new_vel, vel) <= 0.1 * self.adaptivity and self.time_step <= self.max_time_step \
                        and self._magnitude(self, new_vel, vel) <= 0.01 * self.magnitude_adaptivity:
                    # print('Increasing time step. Low deflection wrt velocity')
                    self.streamline.append(new_point)
                    self.svelocity.append(new_vel)
                    self.fvelocity.append(new_fvel)
                    self.time.append(self.time_step)
                    self.point = new_point
                    vel = new_vel.copy()
                    fvel = new_fvel.copy()
                    self.time_step = 2 * self.time_step
                    loop_check = 0
                # Decrease time step when angle is above 1.4 degrees
                # Make sure time step does not go to zero; 1 pico-second
                elif self.angle_btw(new_vel, vel) >= self.adaptivity and self.time_step >= 1e-12 \
                        and self._magnitude(self, new_vel, vel) >= 0.1 * self.magnitude_adaptivity:
                    # print('Decreasing time step. High deflection wrt velocity')
                    self.time_step = 0.5 * self.time_step
                # Save if none of the above conditions meet
                else:
                    self.streamline.append(new_point)
                    self.svelocity.append(new_vel)
                    self.fvelocity.append(new_fvel)
                    self.time.append(self.time_step)
                    self.point = new_point
                    vel = new_vel.copy()
                    fvel = new_fvel.copy()
                    loop_check = 0

            # Save files for each particle; can be used for multiprocessing large number of particles
            self._save_data(self)
            t.stop()

        if method == 'ppath-c-space':
            t = Timer(text="Time taken for particle " + str(self.task) + " is {:.2f} seconds")
            t.start()
            pvel = None
            fvel = None
            save_point = self.point
            idx = Search(grid, self.point)
            idx.compute(method='c-space')
            while True:
                interp = Interpolation(flow, idx)
                interp.compute(method='c-space')
                intg = Integration(interp)
                new_point, new_fvel, new_pvel = intg.compute_ppath(diameter=self.diameter,
                                                                   density=self.density,
                                                                   velocity=pvel, method='cRK4',
                                                                   time_step=self.time_step,
                                                                   drag_model=self.drag_model)
                if new_point is None:
                    # For multi-block case if the point is out-of-block
                    # Use previous point and run one-step of p-space algo
                    print('Point exited the block! Searching for new position...')
                    idx = Search(grid, save_point)
                    interp = Interpolation(flow, idx)
                    intg = Integration(interp)
                    idx.compute(method='p-space')
                    interp.compute()
                    new_point, new_fvel, new_pvel = intg.compute_ppath(diameter=self.diameter,
                                                                       density=self.density,
                                                                       velocity=pvel, method='pRK4',
                                                                       time_step=self.time_step,
                                                                       drag_model=self.drag_model)
                    if new_point is None:
                        print('Point out-of-domain. Integration complete!')
                        break
                    else:
                        # Update the block in idx
                        idx = Search(grid, new_point)
                        idx.compute(method='c-space')
                        self.streamline.append(new_point)
                        self.fvelocity.append(new_fvel)
                        self.svelocity.append(new_pvel)
                        self.time.append(self.time_step)
                        pvel = new_pvel.copy()
                        fvel = new_fvel.copy()
                        self.point = save_point
                        save_point = new_point
                else:

                    # Check for mid-rk4 blowup
                    if intg.rk4_bool is True:
                        print('Mid-RK4 blow up! Reducing time-step')
                        intg.rk4_bool = False
                        self.time_step = 0.5 * self.time_step
                        loop_check += 1
                        if loop_check == 70:
                            print('Stuck in the same loop for too long. Integration ends!')
                            break

                    else:
                        self.point = save_point
                        save_point = idx.c2p(new_point)
                        self.streamline.append(save_point)
                        self.fvelocity.append(new_fvel)
                        self.svelocity.append(new_pvel)
                        self.time.append(self.time_step)
                        pvel = new_pvel.copy()
                        fvel = new_fvel.copy()
                        idx.point = new_point
                        if loop_check > 0:
                            print('Resetting time step')
                            self.time_step = self.time_step / 0.5 ** loop_check
                            loop_check = 0

            # Save files for each particle; can be used for multiprocessing large number of particles
            self._save_data(self)
            t.stop()

        if method == 'adaptive-ppath-c-space':
            t = Timer(text="Time taken for particle " + str(self.task) + " is {:.2f} seconds")
            t.start()
            pvel = None
            fvel = None
            save_point = self.point
            idx = Search(grid, self.point)
            idx.compute(method='c-space')
            while True:
                interp = Interpolation(flow, idx)
                interp.compute(method='c-space')
                intg = Integration(interp)
                new_point, new_fvel, new_pvel = intg.compute_ppath(diameter=self.diameter,
                                                                   density=self.density,
                                                                   velocity=pvel, method='cRK4',
                                                                   time_step=self.time_step,
                                                                   drag_model=self.drag_model)
                # Check if the point is out-of-block
                if new_point is None:
                    # Check for large time-step
                    print('Checking if the end of the domain is reached...')
                    self.time_step = 1e-9 * self.time_step
                    new_point, new_fvel, new_pvel = intg.compute_ppath(diameter=self.diameter,
                                                                       density=self.density,
                                                                       velocity=pvel, method='cRK4',
                                                                       time_step=self.time_step,
                                                                       drag_model=self.drag_model)
                    if new_point is not None:
                        print('Continuing integration by decreasing time-step!')
                        continue
                    # Check for out-of-block situation
                    # For multi-block case if the point is out-of-block
                    # Use previous point and run one-step of p-space algo
                    elif new_point is None:
                        print('Point exited the block! Searching for new position...')
                        idx = Search(grid, save_point)
                        interp = Interpolation(flow, idx)
                        intg = Integration(interp)
                        idx.compute(method='p-space')
                        interp.compute()
                        new_point, new_fvel, new_pvel = intg.compute_ppath(diameter=self.diameter,
                                                                           density=self.density,
                                                                           velocity=pvel, method='pRK4',
                                                                           time_step=self.time_step,
                                                                           drag_model=self.drag_model)
                        if new_point is None:
                            print('Point out-of-domain. Integration complete!')
                            break
                        else:
                            # Update the block in idx
                            idx = Search(grid, new_point)
                            idx.compute(method='c-space')
                            # new_point found is in p-space; so, append
                            self.streamline.append(new_point)
                            self.fvelocity.append(new_fvel)
                            self.svelocity.append(new_pvel)
                            self.time.append(self.time_step)
                            pvel = new_pvel.copy()
                            fvel = new_fvel.copy()
                            self.point = save_point
                            save_point = new_point
                else:
                    self.point = save_point
                    save_point = idx.c2p(new_point)

                    # Check for mid-rk4 blowup
                    if intg.rk4_bool is True:
                        print(f'**WARNING** Large residual. Mid-RK4 blow up! Reducing time-step for particle number'
                              f' - {self.task}')
                        intg.rk4_bool = False
                        self.time_step = 0.5 * self.time_step
                        loop_check += 1
                        if loop_check == 70:
                            print('Stuck in the same loop for too long. Integration ends!')
                            break

                    # Adaptive algorithm starts
                    # Save results and adjust time-step
                    # Details for the algorithm are provided in adaptive-ppath
                    elif pvel is None:
                        # For the first step in the loop
                        self.streamline.append(save_point)
                        self.svelocity.append(new_pvel)
                        self.fvelocity.append(new_fvel)
                        self.time.append(self.time_step)
                        self.point = new_point
                        pvel = new_pvel.copy()
                        fvel = new_fvel.copy()
                    elif self.angle_btw(new_fvel, fvel) is None:
                        print('Increasing time step. Successive points are same')
                        self.time_step = 2 * self.time_step
                        loop_check += 1
                        if loop_check == 70:
                            print(f'Successive points did not change for too long. Integration ends! for particle '
                                  f'{self.task}')
                            break
                    elif self.angle_btw(new_fvel,
                                        fvel) <= 0.1 * self.adaptivity and self.time_step <= self.max_time_step:
                        self.point = save_point
                        save_point = idx.c2p(new_point)
                        self.streamline.append(save_point)
                        self.fvelocity.append(new_fvel)
                        self.svelocity.append(new_pvel)
                        self.time.append(self.time_step)
                        idx.point = new_point.copy()
                        pvel = new_pvel.copy()
                        fvel = new_fvel.copy()
                        self.time_step = 2 * self.time_step
                        loop_check = 0
                    elif self.angle_btw(new_fvel, fvel) >= self.adaptivity and self.time_step >= 1e-12:
                        self.time_step = 0.5 * self.time_step
                    else:
                        self.point = save_point
                        save_point = idx.c2p(new_point)
                        self.streamline.append(save_point)
                        self.fvelocity.append(new_fvel)
                        self.svelocity.append(new_pvel)
                        self.time.append(self.time_step)
                        idx.point = new_point.copy()
                        pvel = new_pvel.copy()
                        fvel = new_fvel.copy()
                        loop_check = 0

            # Save files for each particle; can be used for multiprocessing large number of particles
            self._save_data(self)
            t.stop()

        return



import unittest


class TestDataIO(unittest.TestCase):
    def test_dataio(self):
        from src.io.dataio import DataIO
        from src.io.plot3dio import GridIO, FlowIO
        # grid object
        grid = GridIO('../data/shocks/shock_test.sb.sp.x')
        grid.read_grid()
        grid.compute_metrics()

        # flow object
        flow = FlowIO('../data/shocks/shock_test.sb.sp.q')
        flow.read_flow()

        # data module test
        # data = DataIO(grid, flow, location='../data/shocks/particle_data/multi_process_test/')
        data = DataIO(grid, flow, location='../data/shocks/particle_data/281nm_time_step_adaptive/',
                      read_file='../data/shocks/particle_data/281nm_time_step_adaptive/combined_file.npy')
        # Increased refinement for better resolution
        data.x_refinement = 500
        data.y_refinement = 400
        data.compute()


if __name__ == '__main__':
    unittest.main()



# This file contains DataIO class to read and write particle data

import numpy as np
from src.streamlines.search import Search
from src.streamlines.interpolation import Interpolation
from multiprocessing.pool import ThreadPool as Pool
import multiprocessing as mp
from scipy.interpolate import griddata
import os
import re


class DataIO:
    """
    Use to read/write particle data
    The process happens in four steps. Each step is described below
        1. Reads in the scattered particle data obtained from LPT code. Two ways are available:
            If a combined file is given, data is obtained from that --> faster
            Else, data is read from a folder of files where each file corresponds to a particle --> slow due to for loop
        2. Interpolates fluid data to these scattered locations. Two temp files are generated
            interpolated_q_data.npy --> has the flow data at the scattered locations
            new_p_data.npy --> has the particle data at scattered locations (WITH few outlier points removed)
        3. Lagrangian to Eulerian interpolation. Two temp files are generated
            flow_data.npy --> Fluid data interpolated to the grid
            particle_data.npy --> Particle data interpolated to the grid
        4. mgrid data to plot3d for visualization and syPIV usage. Output tunnel. Three files are generated
            mgrd_to_p3d.x --> Grid file
            mgrd_to_p3d_fluid.q --> fluid data
            mgrd_to_p3d_particle.q --> particle data in Eulerian interpretation

    Attributes
    ----------
    Input :
        grid : src.io.plot3dio.GridIO
            Grid object created from GridIO
        flow : src.io.plot3dio.FlowIO
            Flow object created from FlowIO
        read_file : str
            Combined file of all the particles generated from LPT code
        location : str
            Files location for individual particle data generated from the LPT code
        x_refinement: int
            Refinement of grid in horizontal direction
        y_refinement: int
            Refinement of grid in vertical direction
    Output:
        None

    Methods
    -------
        compute()
            saves the output files in the working directory

        Examples
            The test case is in test/test_dataio.py


    author: Dilip Kalagotla @ kal ~ dilip.kalagotla@gmail.com
    date: 12-28/2022
    """

    def __init__(self, grid, flow, read_file=None, location='.',
                 x_refinement: int = 50, y_refinement: int = 40):
        self.grid = grid
        self.flow = flow
        self.read_file = read_file
        self.location = location
        self.x_refinement = x_refinement
        self.y_refinement = y_refinement

    @staticmethod
    def _natural_sort(_l: list):
        """
        Sorts a list in natural order of things. Humanity is weird
        Args:
            _l: list of strings

        Returns:
            sorted list in natural order

        """
        convert = lambda text: int(text) if text.isdigit() else text.lower()
        alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
        return sorted(_l, key=alphanum_key)

    def compute(self):
        """
        This should interpolate the scattered particle data onto a 2D grid
        Return a file to be used for syPIV
        Returns:
        """
        # read particle files from a folder
        # read particle data
        try:
            print('Trying to read from a combined file...')
            _p_data = np.load(self.read_file)
            print('Read from the combined file!!')
        except:
            print('Reading from a group of files... This will take a while!')
            # Sort in natural order to stack particles in order
            _files = np.array(self._natural_sort(os.listdir(self.location)))
            _bool = []
            for _i, _name in enumerate(_files):
                if not _name.endswith(".npy"):
                    _bool.append(_i)
            _files = np.delete(_files, _bool)

            _p_data = np.load(self.location + _files[0])
            for infile in _files[1:]:
                _temp = np.load(self.location + infile)
                _p_data = np.vstack((_p_data, _temp))
                print('Done with - ' + infile)
            print('Done reading files into an array from a group of files!')
            np.save(self.location + 'combined_file', _p_data)
            print('**SUCCESS** combined_file.npy saved to the given location.')

        _x_min, _x_max = _p_data[:, 0].min(), _p_data[:, 0].max()
        _y_min, _y_max = _p_data[:, 1].min(), _p_data[:, 1].max()

        # Get density and energy for plot3d file at locations
        _locations = _p_data[:, :3]

        def _flow_data(_point, _index, _size):
            """
            Internal function that interpolates data to scattered points
            Args:
                _point: One of the scattered points

            Returns:
                Interpolated data at each scattered point location given

            """
            _idx = Search(self.grid, _point)
            _idx.compute(method='p-space')
            _interp = Interpolation(self.flow, _idx)
            _interp.compute(method='p-space')

            try:
                print(f'Done with {_index}/{_size}')
                return _interp.q.reshape(-1)
            except:
                print('Returned None. Exception occurred')
                return np.array([1], dtype=int)

        try:
            # Read back the saved file
            _q_list = np.load(self.location + 'dataio/interpolated_q_data.npy', allow_pickle=False)
            _p_data = np.load(self.location + 'dataio/new_p_data.npy', allow_pickle=False)
            print('Read the available interpolated data to continue with the griddata algorithm')
        except:
            print('Interpolated data file is unavailable. Continuing with interpolation to scattered data!\n'
                  'This is going to take sometime. Sit back and relax!\n'
                  'Your PC will take off because of multi-process. Let it breathe...\n')
            _processors = mp.cpu_count()
            _pool = Pool(_processors)
            _loc_len = len(_locations)
            # Passing extra parameters to keep track of the progress. Chunk-size helps to keep it orderly
            _q_list = _pool.starmap(_flow_data, zip(_locations, np.arange(0, _loc_len), np.repeat(_loc_len, _loc_len)),
                                    chunksize=1)
            _pool.close()

            # Fluid data at scattered points/particle locations
            # Some searches return None. This helps remove those locations!
            _remove_index = [j for j in range(len(_q_list)) if np.all(_q_list[j] == 1)]
            _q_list = np.vstack(np.delete(_q_list, _remove_index, axis=0))
            _p_data = np.delete(_p_data, _remove_index, axis=0)
            # _q_list = np.array([i for i in _q_list if np.all(i != 1)])
            # Save both interpolated data and new particle data for easy future computations
            try:
                # Try creating the directory; if exists errors out and except
                os.mkdir(self.location + 'dataio')
                np.save(self.location + 'dataio/interpolated_q_data', _q_list)
                np.save(self.location + 'dataio/new_p_data', _p_data)
                print('Created dataio and saved interpolated flow data to scattered points.\n')
            except:
                np.save(self.location + 'dataio/interpolated_q_data', _q_list)
                np.save(self.location + 'dataio/new_p_data', _p_data)
            print('Done with interpolating flow data to scattered points.\n')

        # Particle data at the scattered points/particle locations
        # rho, x,y,z - momentum, energy per unit volume (q-file data)
        _q_p_list = np.hstack((_q_list[:, 0].reshape(-1, 1), _p_data[:, 3:6] * _q_list[:, 0].reshape(-1, 1),
                               _q_list[:, 4].reshape(-1, 1)))
        # Fluid data at the scattered points/particle locations
        _q_f_list = np.hstack((_q_list[:, 0].reshape(-1, 1), _p_data[:, 6:9] * _q_list[:, 0].reshape(-1, 1),
                               _q_list[:, 4].reshape(-1, 1)))

        # Create the grid to interpolate to
        _xi, _yi = np.linspace(_x_min, _x_max, self.x_refinement), np.linspace(_y_min, _y_max, self.y_refinement)
        _xi, _yi = np.meshgrid(_xi, _yi, indexing='ij')

        # Function to loop through the scattered data
        def _grid_interp(_data=None, _points=_p_data[:, :2], _x_grid=_xi, _y_grid=_yi, method='linear'):
            """
            Interpolate to grid data from scatter points
            Args:
                _points: scattered data - This is fixed, usually
                _x_grid: grid data - fixed usually
                _y_grid: grid data - fixed usually
                _data: data set to be interpolated
                method: set to linear but can be changed if needed

            Returns:
                Interpolated flow data

            """
            _data = _data.reshape(-1)
            # Transposing to keep consistency with default xy indexing of meshgrid
            _q = griddata(_points, _data, (_x_grid, _y_grid), method=method, fill_value=_data.max() * 2)

            return _q

        try:
            # Read to see if data is available
            _qf = np.load(self.location + 'dataio/flow_data.npy')
            _qp = np.load(self.location + 'dataio/particle_data.npy')
            print('Loaded available flow/particle data from numpy residual files\n')
        except:
            print('Interpolating data to the grid provided...\n')
            # Interpolate scattered data onto the grid -- for flow
            _pool = Pool(mp.cpu_count())
            _qf = _pool.map(_grid_interp,
                            [_q_f_list[:, 0], _q_f_list[:, 1], _q_f_list[:, 2], _q_f_list[:, 3], _q_f_list[:, 4]],
                            chunksize=1)
            _pool.close()
            print(f'Done with flow data interpolation to grid.\n')

            # Save the array to a file
            # This will only happen when there are files in dataio directory
            _qf = np.array(_qf)
            np.save(self.location + 'dataio/flow_data', _qf)

            # Interpolate scattered data onto the grid -- for particles
            _pool = Pool(mp.cpu_count())
            _qp_123 = _pool.map(_grid_interp, [_q_p_list[:, 1], _q_p_list[:, 2], _q_p_list[:, 3]], chunksize=1)
            _pool.close()
            print(f'Done with particle data interpolation to grid.\n')

            # Create _qp array from known values
            _qp = np.dstack((_qf[0], _qp_123[0], _qp_123[1], _qp_123[2], _qf[-1])).transpose((2, 0, 1))

            # Save data to a temporary file
            _qp = np.array(_qp)

            # This will only happen when there are files in dataio directory
            np.save(self.location + 'dataio/particle_data', _qp)

        # Write out to plot3d format for further processing
        self.grid.mgrd_to_p3d(_xi, _yi, out_file=self.location + 'dataio/mgrd_to_p3d.x')
        self.flow.mgrd_to_p3d(_qf, mode='fluid', out_file=self.location + 'dataio/mgrd_to_p3d')
        self.flow.mgrd_to_p3d(_qp, mode='particle', out_file=self.location + 'dataio/mgrd_to_p3d')
        print('Files written to the working directory')

        return



# Use tri-linear interpolation to get data at the given point

import numpy as np


class Interpolation:
    """
    Module to do the cell data interpolation to the given point

    ...

    Attributes
    ----------
    Input:
        flow : src.io.plot3dio.FlowIO
            Flow object created from FlowIO
        idx: src.streamlines.Search
            Index object created from Search
    Output:
        q : ndarray
            Interpolated data at the given point
        nb, ni, nj, nk: int
            Dimensionless data from flow
        mach, alpha, rey, time: float
            Dimensionless data from flow
        J_inv: ndarray
            Inverse grid metrics at the given point

    Methods
    -------
    compute(method='p-space')
        Interpolates the data onto a given point
        p-space or c-space can be specified

    Example:
        grid = GridIO('../data/plate_data/plate.sp.x')
        flow = FlowIO('../data/plate_data/sol-0000010.q')
        idx = Search(grid, [8.5, 0.5, 0.01])
        interp_data = Interpolation(grid, flow, idx)

        grid.read_grid()
        flow.read_flow()
        idx.compute()
        interp_data.compute()  # method='p-space' is default

    author: Dilip Kalagotla @ kal ~ dilip.kalagotla@gmail.com
    date: 10-29/2021
    """

    def __init__(self, flow, idx):
        self.flow = flow
        self.idx = idx
        self.q = None
        self.nb = None
        self.ni, self.nj, self.nk = [None] * 3
        self.mach, self.alpha, self.rey, self.time = [None] * 4
        self.J = None
        self.J_inv = None

    def __str__(self):
        doc = "This instance uses " + self.flow.filename + " as the flow file " \
                                                           "to compute properties at " + self.idx.ppoint + "\n"
        return doc

    def compute(self, method='p-space'):
        """
        Find interpolated plot3d data and grid metrics at a given point
        Args:
            method: Available methods for interpolation
            1. p-space: Tri-linear interpolation in physical co-ordinates
            2. c-space: Tri-linear interpolation in c-space. Must-use for the c-space algo

        Returns:
            q --> Attribute with the interpolated flow data.
            Has same ndim as q attribute from flow object

        """

        # Assign data from q file to keep the format for further computations
        self.nb = self.flow.nb
        self.ni, self.nj, self.nk = self.flow.ni, self.flow.nj, self.flow.nk
        self.mach, self.alpha, self.rey, self.time = self.flow.mach, self.flow.alpha, self.flow.rey, self.flow.time

        # If out of domain return
        if self.idx.ppoint is None and self.idx.cpoint is None:
            print('Cannot run interpolation. The given point is out of domain.\n')
            self.q = None
            return

        # Obtain grid data from nodes found from the search method
        _cell_grd = np.zeros(self.idx.cell.shape)
        _cell_grd = self.idx.grid.grd[self.idx.cell[:, 0], self.idx.cell[:, 1], self.idx.cell[:, 2], :, self.idx.block]

        # Obtain flow data from nodes found
        _cell_q = np.zeros((self.idx.cell.shape[0], 5))
        _cell_q = self.flow.q[self.idx.cell[:, 0], self.idx.cell[:, 1], self.idx.cell[:, 2], :, self.idx.block]

        match method:
            # "Tri"Linear interpolation
            case 'p-space':
                # If the node is found in a cell
                if self.idx.cell.shape == (8, 3) and self.idx.info is None:
                    def _f(self, _m, _n, _arr1, _arr2, _axis, _paxis, _data1, _data2):
                        """
                        Internal function used for linearly interpolating data
                        :param _m: int
                            point-1 for interpolation
                        :param _n: int
                            point-2 for interpolation
                        :param _arr1: ndarray
                            data with point-1
                        :param _arr2: ndarray
                            data with point-2
                        :param _axis: int
                            assign x, y or z axis to consider for interpolation
                        :param _paxis: int
                            assign point's x, y, or z for interpolation
                        :param _data1: ndarray/list
                            data at point-1 to be interpolated. List is provided when moving across axis
                        :param _data2: ndarray/list
                            data at point-2 to be interpolated. List is provided when moving across axis
                        :return: None

                        author: Dilip Kalagotla @ kal ~ dilip.kalagotla@gmail.com
                        date: 10-29/2021
                        """
                        _a = (_arr2[_n, _axis] - self.idx.ppoint[_paxis]) / (_arr2[_n, _axis] - _arr1[_m, _axis]) * _data1[_m]
                        _b = (self.idx.ppoint[_paxis] - _arr1[_m, _axis]) / (_arr2[_n, _axis] - _arr1[_m, _axis]) * _data2[_n]
                        return _a + _b

                    # 0123 are vertices of quadrilateral
                    # Doing interpolation to x location in self.idx.ppoint
                    # Values in the function help set the linear interpolation in x-direction
                    _cell_grd_01 = _f(self, 0, 1, _cell_grd, _cell_grd, 0, 0, _cell_grd, _cell_grd)
                    _cell_grd_32 = _f(self, 3, 2, _cell_grd, _cell_grd, 0, 0, _cell_grd, _cell_grd)
                    _cell_q_01 = _f(self, 0, 1, _cell_grd, _cell_grd, 0, 0, _cell_q, _cell_q)
                    _cell_q_32 = _f(self, 3, 2, _cell_grd, _cell_grd, 0, 0, _cell_q, _cell_q)

                    # Doing interpolation to the point projection on the face
                    _cell_grd_0123 = _f(self, 1, 1, _cell_grd_01, _cell_grd_32, None, 1,
                                        [None, _cell_grd_01], [None, _cell_grd_32])
                    _cell_q_0123 = _f(self, 1, 1, _cell_grd_01, _cell_grd_32, None, 1,
                                      [None, _cell_q_01], [None, _cell_q_32])

                    # 4567 face
                    # Doing interpolation to x location in self.idx.ppoint
                    # Values in the function help set the linear interpolation in y-direction
                    _cell_grd_45 = _f(self, 4, 5, _cell_grd, _cell_grd, 0, 0, _cell_grd, _cell_grd)
                    _cell_grd_76 = _f(self, 7, 6, _cell_grd, _cell_grd, 0, 0, _cell_grd, _cell_grd)
                    _cell_q_45 = _f(self, 4, 5, _cell_grd, _cell_grd, 0, 0, _cell_q, _cell_q)
                    _cell_q_76 = _f(self, 7, 6, _cell_grd, _cell_grd, 0, 0, _cell_q, _cell_q)

                    # Doing interpolation to the point projection on the face
                    _cell_grd_4567 = _f(self, 1, 1, _cell_grd_45, _cell_grd_76, None, 1, [None, _cell_grd_45],
                                        [None, _cell_grd_76])
                    _cell_q_4567 = _f(self, 1, 1, _cell_grd_45, _cell_grd_76, None, 1, [None, _cell_q_45],
                                      [None, _cell_q_76])

                    # Doing data interpolation to the given point based on the face points
                    self.q = _f(self, 2, 2, _cell_grd_0123, _cell_grd_4567, None, 2,
                                [None, None, _cell_q_0123], [None, None, _cell_q_4567])
                    self.q = self.q.reshape((1, 1, 1, -1, 1))

                # if the point is node return node data
                if self.idx.info == 'Given point is a node in the domain with a tol of 1e-6.\n' \
                                    'Interpolation will assign node properties for integration.\n'\
                                    'Index of the node will be returned by cell attribute\n':
                    self.q = _cell_q
                    self.q = self.q.reshape((1, 1, 1, -1, 1))

            case 'c-space':
                """
                Point from idx is in c-space. Tri-linear interpolation is performed.
                Equation can be found in Sadarjoen et al.
                """

                # if the point is node return node data
                if self.idx.info == 'Given point is a node in the domain with a tol of 1e-6.\n' \
                                    'Interpolation will assign node properties for integration.\n' \
                                    'Index of the node will be returned by cell attribute\n':
                    self.q = _cell_q
                    self.q = self.q.reshape((1, 1, 1, -1, 1))
                    return

                # Interpolate m1 and m2 for use in integration; m1 is used in ppath only!
                # m1 -- indicated as J (determinant of m1 is J)
                _cell_J = self.idx.grid.m1[self.idx.cell[:, 0], self.idx.cell[:, 1], self.idx.cell[:, 2], :, :, self.idx.block]
                # m2 -- indicated as J_inv (determinant of m2 is J_inv)
                _cell_J_inv = self.idx.grid.m2[self.idx.cell[:, 0], self.idx.cell[:, 1], self.idx.cell[:, 2], :, :, self.idx.block]
                _alpha, _beta, _gamma = self.idx.cpoint - self.idx.cell[0]

                def _eqn(_alpha, _beta, _gamma, _var):
                    _fun = (1 - _alpha) * (1 - _beta) * (1 - _gamma) * _var[0] + \
                                _alpha  * (1 - _beta) * (1 - _gamma) * _var[1] + \
                                _alpha  *      _beta  * (1 - _gamma) * _var[2] + \
                           (1 - _alpha) *      _beta  * (1 - _gamma) * _var[3] + \
                           (1 - _alpha) * (1 - _beta) *      _gamma  * _var[4] + \
                                _alpha  * (1 - _beta) *      _gamma  * _var[5] + \
                                _alpha  *      _beta *       _gamma  * _var[6] + \
                           (1 - _alpha) *      _beta  *      _gamma  * _var[7]

                    return _fun

                self.q = _eqn(_alpha, _beta, _gamma, _cell_q)
                self.J = _eqn(_alpha, _beta, _gamma, _cell_J)
                self.J_inv = _eqn(_alpha, _beta, _gamma, _cell_J_inv)

                self.q = self.q.reshape((1, 1, 1, -1, 1))

            case 'rbf-p-space':
                """
                Raidal basis function interpolation in physical space
                """
                # if the point is node return node data
                if self.idx.info == 'Given point is a node in the domain with a tol of 1e-6.\n' \
                                    'Interpolation will assign node properties for integration.\n' \
                                    'Index of the node will be returned by cell attribute\n':
                    self.q = _cell_q
                    self.q = self.q.reshape((1, 1, 1, -1, 1))
                    return

                from scipy.interpolate import RBFInterpolator as rbf
                _rbf = rbf(_cell_grd, _cell_q)
                self.q = _rbf(np.array(self.idx.ppoint).reshape(1, -1))
                self.q = self.q.reshape((1, 1, 1, -1, 1))

            case 'rbf-c-space':
                """
                Raidal basis function interpolation in c-space
                """
                # if the point is node return node data
                if self.idx.info == 'Given point is a node in the domain with a tol of 1e-6.\n' \
                                    'Interpolation will assign node properties for integration.\n' \
                                    'Index of the node will be returned by cell attribute\n':
                    self.q = _cell_q
                    self.q = self.q.reshape((1, 1, 1, -1, 1))
                    return

                # Interpolate m1 and m2 for use in integration; m1 is used in ppath only!
                # m1 -- indicated as J (determinant of m1 is J)
                _cell_J = self.idx.grid.m1[self.idx.cell[:, 0], self.idx.cell[:, 1], self.idx.cell[:, 2], :, :,
                          self.idx.block]
                # m2 -- indicated as J_inv (determinant of m2 is J_inv)
                _cell_J_inv = self.idx.grid.m2[self.idx.cell[:, 0], self.idx.cell[:, 1], self.idx.cell[:, 2], :, :,
                              self.idx.block]
                # _alpha, _beta, _gamma and reshape for rbf intepolator to work
                _fractions = (self.idx.cpoint - self.idx.cell[0]).reshape(1, -1)

                # Start RBF interpolation
                from scipy.interpolate import RBFInterpolator as rbf
                _unit_cell = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0],
                                       [0, 1, 0], [0, 0, 1], [1, 0, 1],
                                       [1, 1, 1], [0, 1, 1]])
                _rbf_q = rbf(_unit_cell, _cell_q)
                self.q = _rbf_q(_fractions)
                self.q = self.q.reshape((1, 1, 1, -1, 1))

                _rbf_J = rbf(_unit_cell, _cell_J)
                self.J = _rbf_J(_fractions)
                _rbf_J_inv = rbf(_unit_cell, _cell_J_inv)
                self.J_inv = _rbf_J_inv(_fractions)





# import all the streamlines modules
from .search import Search
from .interpolation import Interpolation
from .integration import Integration
from .streamlines import Streamlines
from .stochastic_model import StochasticModel, Particle, SpawnLocations



# Class to run streamlines script in parallel
# Stochastic model for tracers is implemented

import numpy as np
from multiprocessing.pool import ThreadPool as Pool
import multiprocessing as mp
from src.streamlines.streamlines import Streamlines

rng = np.random.default_rng(7)


class StochasticModel(Streamlines):
    """Module to spawn and run LPT on given tracers parallely

    ...

    Attributes
    ----------

    """

    def __init__(self, particles, spawn_locations, method='adaptive-p-space',
                 grid=None, flow=None, point=None,
                 search='p-space', interpolation='p-space', integration='pRK4',
                 diameter=1e-7, density=1000,
                 time_step=1e-3, max_time_step=1, drag_model='henderson', adaptivity=0.001,
                 magnitude_adaptivity=0.001,
                 filepath: str = None
                 ):
        super().__init__(point=point,
                         search=search, interpolation=interpolation, integration=integration,
                         diameter=diameter, density=density,
                         time_step=time_step, max_time_step=max_time_step, drag_model=drag_model)
        self.particles = particles
        self.spawn_locations = spawn_locations
        # Read-in grid and flow files
        self.grid = grid
        self.flow = flow
        self.method = method
        self.filepath = filepath
        self.adaptivity = adaptivity
        self.magnitude_adaptivity = magnitude_adaptivity

    def setup(self, spawn_location, particle_dia, task):
        """
        Sets up the function to be run in parallel
        Args:
            self:
            spawn_location:
            particle_dia:
            task: same as particle.n_concentration, used to track progress of computation

        Returns:

        """
        # TODO: Have to use inheritance properties. Currently, just calling in another object
        print(f'Execution started for particle number - {task}')
        sl = Streamlines(None, None, point=spawn_location, diameter=particle_dia, time_step=self.time_step,
                         task=task)
        sl.density = self.particles.density
        sl.drag_model = self.drag_model
        sl.max_time_step = self.max_time_step
        sl.filepath = self.filepath
        sl.search = self.search
        sl.interpolation = self.interpolation
        sl.integration = self.integration
        sl.adaptivity = self.adaptivity
        sl.magnitude_adaptivity = self.magnitude_adaptivity
        sl.compute(method=self.method, grid=self.grid, flow=self.flow)

        return sl

    def multi_process(self):
        """
        To parallelize using multiprocessing approach; the setup function
        Returns:

        """
        with mp.Pool(mp.cpu_count() - 1) as pool:
            lpt_data = pool.starmap(self.setup, zip(self.spawn_locations.locations, self.particles.particle_field,
                                                    np.arange(self.particles.n_concentration)), chunksize=1)

        return lpt_data

    def multi_thread(self):
        """
        To parallelize using multithreading approach; the setup function
        Returns:

        """
        with Pool(mp.cpu_count() - 1) as pool:
            lpt_data = pool.starmap(self.setup, zip(self.spawn_locations.locations, self.particles.particle_field,
                                                    np.arange(self.particles.n_concentration)), chunksize=1)

        return lpt_data

    def serial(self):
        """
        To run setup in serial
        Returns:

        """
        # Run setup function in serial
        lpt_data = []
        for i in range(self.particles.n_concentration):
            lpt_data.append(self.setup(self.spawn_locations.locations[i], self.particles.particle_field[i], i))

        return lpt_data


class Particle:
    """
    Class holds details for particles used in a PIV experiment
    ---
    User has to provide all the information to generate size distribution
    """

    def __init__(self):
        self.distribution = "gaussian"
        self.min_dia = None
        self.max_dia = None
        self.mean_dia = None
        self.std_dia = None
        self.density = None
        self.n_concentration = None
        self.particle_field = None

    def compute_distribution(self):
        """
        Run this method to return a distribution of particle diameters
        :return: numpy.ndarray
        A 1d array of particle diameters
        """
        if self.distribution == "gaussian":
            print("When Gaussian distribution is used,"
                  " the particle statistics are computed using mean and std diameters\n"
                  "Particle min and max are cutoffs for the distribution")
            self.particle_field = rng.normal(self.mean_dia, self.std_dia, int(self.n_concentration))
            self.particle_field = np.clip(self.particle_field, self.min_dia, self.max_dia)
            np.random.shuffle(self.particle_field)
            return

        # TODO: Add Uniform distribution

    pass


class SpawnLocations:
    """
    Creates spawn locations array based on number of particles
    """
    def __init__(self, particles):
        self.x_min, self.x_max = None, None
        self.y_min, self.y_max = None, None
        self.z_min, self.z_max = None, None
        self.particles = particles
        self.locations = None

    def compute(self):
        """
        Computes the locations array to be passed into parallel
        Returns:

        """
        _size = self.particles.n_concentration
        # Draw a straight line between given points
        if self.x_max is None and self.z_max is None:
            _x_temp = np.repeat(self.x_min, _size).reshape(_size, 1)
            _z_temp = np.repeat(self.z_min, _size).reshape(_size, 1)
            _y_temp = np.linspace(self.y_min, self.y_max, _size).reshape(_size, 1)

            self.locations = np.hstack((_x_temp, _y_temp, _z_temp))



# This test is for the search algorithm using plate plot3d data

import unittest
from parameterized import parameterized
from src.function.timer import Timer


# This setup only works with the plate data
# Need to add assertions to parametrized and change them to variables in the function
# to be able to implement
class TestSearch(unittest.TestCase):
    # sb: single block; mb: multi block
    # sp: single precision; dp: double precision
    # Add more cases as a tuple below
    @parameterized.expand([
        ('sb_sp_distance', '../data/plate_data/plate.sp.x', 'f4', 'distance'),
        ('sb_sp_block_distance', '../data/plate_data/plate.sp.x', 'f4', 'block_distance'),
        ('mb_dp_distance', '../data/multi_block/plate/plate.mb.dp.x', 'f8', 'distance'),
        ('mb_dp_block_distance', '../data/multi_block/plate/plate.mb.dp.x', 'f8', 'block_distance'),
        ('mb_dp_block_distance', '../data/multi_block/plate/plate.mb.dp.x', 'f8', 'c-space')

    ])
    @Timer()
    def test_search(self, name, filename='../data/plate_data/plate.sp.x', data_type='f4', method='binary'):
        from src.io.plot3dio import GridIO
        from src.streamlines.search import Search

        # Read the grid data
        grid = GridIO(filename)
        grid.read_grid(data_type=data_type)
        grid.compute_metrics()
        flow = None

        # Start test for search algorithm
        # Check if distance and neighbors are working fine
        idx = Search(grid, [8.5, 0.5, 0.01])
        idx.compute(method=method)
        self.assertEqual(idx.cell.shape, (8, 3))
        self.assertEqual(idx.info, None)
        # Check for out of domain point
        idx = Search(grid, [-10, 0.5, 0.01])
        idx.compute(method=method)
        self.assertEqual(idx.cell, None)
        self.assertEqual(idx.info,
                         'Given point is not in the domain. The cell attribute will return "None"'
                         ' in search algorithm\n')
        # Check for on the node case
        # Used a Node from grid numbering
        idx = Search(grid, grid.grd[0, 0, 0, :, 0])
        idx.compute(method='block_distance')
        self.assertEqual(idx.info, 'Given point is a node in the domain with a tol of 1e-6.\n'
                                   'Interpolation will assign node properties for integration.\n'
                                   'Index of the node will be returned by cell attribute\n')
        # Check for on the boundary case
        # Point is created by averaging neighboring nodes
        idx = Search(grid, [sum(grid.grd[0:2, 0, 0, 0, 0]) / 2, grid.grd[0, 0, 0, 1, 0], grid.grd[0, 0, 0, 2, 0]])
        idx.compute(method=method)
        self.assertEqual(idx.cell.shape, (8, 3))
        self.assertEqual(idx.info, None)

    def test_p2c(self, name='p2c_test', filename='../data/plate_data/plate.sp.x', data_type='f4'):
        from src.io.plot3dio import GridIO
        from src.streamlines.search import Search

        # Read the grid data
        grid = GridIO(filename)
        grid.read_grid(data_type=data_type)
        grid.compute_metrics()
        flow = None

        # Start test for search algorithm
        # Check if distance and neighbors are working fine
        point = [8.5, 0.5, 0.01]
        idx = Search(grid, point)
        idx.p2c(point)


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np


class TestObliqueShockResponse(unittest.TestCase):
    def test_oblique_shock(self):
        from src.test_cases.oblique_shock_data import ObliqueShock
        os = ObliqueShock()
        os.mach = 2.3
        os.deflection = 10
        os.compute()
        self.assertAlmostEqual(os.shock_angle.all(), np.array([34.32642717, 85.02615188]).all(), places=4)

    def test_oblique_shock_relation(self):
        from src.test_cases.oblique_shock_data import ObliqueShock
        from src.test_cases.oblique_shock_data import ObliqueShockData
        os = ObliqueShock()
        os.mach = 2.3
        os.deflection = 10
        os.compute()

        # Create grid and flow files
        osd = ObliqueShockData()
        osd.nx_max = 10
        osd.ny_max = 10
        osd.nz_max = 10
        osd.inlet_density = 1.273
        osd.inlet_temperature = 300
        osd.inlet_pressure = 101325
        osd.points = 100
        osd.oblique_shock = os
        osd.create_grid()
        osd.create_flow()


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

VERSION = "0.1"
DESCRIPTION = "A Lagrangian Particle Tracking package"
LONG_DESCRIPTION = "Includes a set of tools for Lagrangian Particle Tracking like search, interpolation, etc."

# Setting up
setup(
    # name must match the folder name
    name="project-arrakis",
    version=VERSION,
    author="kal @ Dilip Kalagotla",
    author_email="<dilipkalagotla@gmail.com>",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=[],  # add any additional packages that
    # needs to be installed along with your package. Eg: 'caer'
    keywords=["python", "first package"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
)



import unittest


class TestVariables(unittest.TestCase):
    from src.io.plot3dio import GridIO, FlowIO
    from src.streamlines.search import Search
    from src.streamlines.interpolation import Interpolation

    # Get the data at a given point
    grid = GridIO('../data/plate_data/plate.sp.x')
    flow = FlowIO('../data/plate_data/sol-0000010.q')
    idx = Search(grid, [8.5, 0.5, 0.01])
    point_data = Interpolation(flow, idx)

    grid.read_grid()
    flow.read_flow()
    idx.compute()
    point_data.compute()

    def test_variables(self, flow=flow):
        """
        To test variables class for the whole domain
        Test each function separately
        :param flow: FlowIO object
        :return: None
        """
        from src.function.variables import Variables

        variables = Variables(flow)
        variables.compute_velocity()
        variables.compute_temperature()

        self.assertEqual(variables.velocity.shape, (720, 152, 129, 3, 1))
        self.assertEqual(variables.temperature.shape, (720, 152, 129, 1))

    def test_variables_compute(self, flow=flow):
        """
        Test "compute" method in variables class
        :param flow: FlowIO object
        :return: None
        """
        from src.function.variables import Variables

        variables = Variables(flow)
        variables.compute()

        self.assertEqual(variables.velocity.shape, (720, 152, 129, 3, 1))
        self.assertEqual(variables.temperature.shape, (720, 152, 129, 1))

    def _test(self, point_variables):
        """
        Inside function to test point_variables

        :param point_variables: Interpolation object
        :return: None
        """

        self.assertEqual(
            sum(abs(point_variables.velocity.reshape(3) - [1.02420611e-01, -5.38896289e-06, 6.40980361e-09])) <= 1e-6,
            True)
        self.assertEqual(abs(point_variables.temperature.reshape(1) - 0.97452141) <= 1e-6, True)

    def test_point_variables(self, point_data=point_data):
        """
        Test variables class for a single point
        :param point_data: Interpolation object
        :return: None
        """
        from src.function.variables import Variables

        point_variables = Variables(point_data)
        point_variables.compute_velocity()
        point_variables.compute_temperature()

        self._test(point_variables)

    def test_point_variables_compute(self, point_data=point_data):
        """
        Test "compute" method in variables class for a single point
        :param point_data: Interpolation object
        :return: None
        """
        from src.function.variables import Variables

        point_variables = Variables(point_data)
        point_variables.compute()

        self._test(point_variables)


if __name__ == '__main__':
    unittest.main()



import unittest
from parameterized import parameterized
from src.function.timer import Timer


class TestIntegration(unittest.TestCase):
    @parameterized.expand([
        ('sb_sp_p_space', '../data/plate_data/plate.sp.x', '../data/plate_data/sol-0000010.q',
         'f4', [8.5, 0.5, 0.01], 'block_distance', 'p-space', 'p-space'),
        ('sb_sp_c_space', '../data/plate_data/plate.sp.x', '../data/plate_data/sol-0000010.q',
         'f4', [8.5, 0.5, 0.01], 'c-space', 'c-space', 'c-space'),
        ('sb_sp_p_space', '../data/plate_data/plate.sp.x', '../data/plate_data/sol-0000010.q',
         'f4', [8.5, 0.5, 0.01], 'block_distance', 'p-space', 'RK4'),
        ('sb_sp_c_space', '../data/plate_data/plate.sp.x', '../data/plate_data/sol-0000010.q',
         'f4', [8.5, 0.5, 0.01], 'c-space', 'c-space', 'cRK4'),
        ('mb_sp_p_space', '../data/multi_block/plate/plate.mb.sp.x', '../data/multi_block/plate/plate.mb.sp.q',
         'f4', [8.5, 0.5, 0.01], 'block_distance', 'p-space', 'RK4'),
        ('mb_sp_c_space', '../data/multi_block/plate/plate.mb.sp.x', '../data/multi_block/plate/plate.mb.sp.q',
         'f4', [8.5, 0.5, 0.01], 'c-space', 'c-space', 'cRK4')

    ])
    @Timer()
    def test_integration(self, name, gridfile='../data/plate_data/plate.sp.x',
                         flowfile='../data/plate_data/sol-0000010.q', data_type='f4', point=None,
                         search_method='block_distance', interpolation_method='p-space', integration_method='RK4'):

        from src.io.plot3dio import GridIO, FlowIO
        from src.streamlines.search import Search
        from src.streamlines.interpolation import Interpolation
        from src.streamlines.integration import Integration

        if point is None:
            point = [8.5, 0.5, 0.01]

        # Read the grid data
        grid = GridIO(gridfile)
        grid.read_grid(data_type=data_type)
        grid.compute_metrics()

        # Read the flow data
        flow = FlowIO(flowfile)
        flow.read_flow(data_type=data_type)

        # Search for the given point
        idx = Search(grid, point)
        idx.compute(method=search_method)

        # Do Interpolation
        interp = Interpolation(flow, idx)
        interp.compute(method=interpolation_method)

        # Do Integration
        intg = Integration(interp)
        new_point = intg.compute(method=integration_method, time_step=1e-2)


if __name__ == '__main__':
    unittest.main()



# Class to calculate particle response across an oblique shock

import numpy as np
from scipy.optimize import newton
from src.streamlines.stochastic_model import StochasticModel, Particle, SpawnLocations
import matplotlib.pyplot as plt
import os
import glob


# Create a class to calculate oblique shock properties from given mach and deflection angle
class ObliqueShock:
    """
    Class to calculate oblique shock properties from given mach and deflection angle

    ...

    Attributes
    ----------
    Input :
        mach : float
            Mach number
        deflection : float
            Deflection angle in degrees
        shock_angle : float
            Shock angle in degrees
    Output :
        pressure_ratio : float
            Pressure ratio across the shock
        density_ratio : float
            Density ratio across the shock
        temperature_ratio : float
            Temperature ratio across the shock


    Methods
    -------
    oblique_shock_relation(_mach, _deflection, _shock_angle)
        Oblique shock relations for given mach, deflection and theta
    derivative(f, method='central', h=1e-7) -- static method
        Compute the difference formula for f with step size h.
    compute()
        Calculate the oblique shock properties

    Example:
    >>> os = ObliqueShock()
    >>> os.mach = 2.3
    >>> os.deflection = 10
    >>> os.compute()
    >>> print(os.shock_angle)
    [34.32642717 85.02615188]

    """

    def __init__(self, mach=None, deflection=None, shock_angle=None):
        self.mach = mach
        self.deflection = deflection
        self.shock_angle = shock_angle
        self.pressure_ratio = None
        self.density_ratio = None
        self.temperature_ratio = None
        self.mach_ratio = None
        self.gamma = 1.4
        self.gas_constant = 287.058

    def oblique_shock_relation(self, _mach, _deflection, _shock_angle):
        """
        Oblique shock relations for given mach, deflection and theta
        Args:

        Returns:
            delta-theta-mach relation
        """
        return (_mach ** 2 * (self.gamma + np.cos(2 * _shock_angle)) + 2) * np.tan(_deflection) * \
            np.tan(_shock_angle) - 2 * (_mach ** 2 * np.sin(_shock_angle) ** 2 - 1)

    @staticmethod
    def derivative(f, method='central', h=1e-7):
        """Compute the difference formula for f with step size h.

        Parameters
        ----------
        f : function
            Vectorized function of one variable
        method : string
            Difference formula: 'forward', 'backward' or 'central'
        h : number
            Step size in difference formula

        Returns
        -------
        function
            Difference formula:
                central: f(x+h) - f(x-h))/2h
                forward: f(x+h) - f(x))/h
                backward: f(x) - f(x-h))/h
        """

        if method == 'central':
            return lambda x: (f(x + h) - f(x - h)) / (2 * h)
        elif method == 'forward':
            return lambda x: (f(x + h) - f(x)) / h
        elif method == 'backward':
            return lambda x: (f(x) - f(x - h)) / h
        else:
            raise ValueError("Method must be 'central', 'forward' or 'backward'.")

    def compute(self):
        # Use Newton-Raphson method to calculate the shock angle
        # Compute the shock angle if mach and deflection are given
        if self.mach is not None and self.deflection is not None:
            self.deflection = np.radians(self.deflection)

            f_shock_angle = lambda shock_angle: self.oblique_shock_relation(self.mach, self.deflection, shock_angle)
            df_shock_angle = self.derivative(f_shock_angle)

            self.shock_angle = newton(f_shock_angle, (0.5, 1.57), fprime=df_shock_angle)

        # Calculate the pressure ratio
        self.pressure_ratio = 1 + 2 * self.gamma / (self.gamma + 1) * (
                self.mach ** 2 * np.sin(self.shock_angle) ** 2 - 1)

        # Calculate the density ratio
        self.density_ratio = (self.gamma + 1) * self.mach ** 2 * np.sin(self.shock_angle) ** 2 / \
                             (2 + (self.gamma - 1) * self.mach ** 2 * np.sin(self.shock_angle) ** 2)

        # Calculate the temperature ratio
        self.temperature_ratio = self.pressure_ratio / self.density_ratio

        # Calculate the mach ratio
        self.mach_ratio = 1/self.mach * np.sqrt(
            (1 + (self.gamma - 1) / 2 * self.mach ** 2 * np.sin(self.shock_angle) ** 2) /
            (self.gamma * self.mach ** 2 * np.sin(self.shock_angle) ** 2 -
             (self.gamma - 1) / 2)) / np.sin(self.shock_angle - self.deflection)

        # Change the shock angle to degrees
        self.shock_angle = np.degrees(self.shock_angle)
        self.deflection = np.degrees(self.deflection)

        return


# class to create a grid and flow based on oblique shock properties
class ObliqueShockData:
    """
    Class to create a grid and flow based on oblique shock properties
    """
    def __init__(self, oblique_shock=ObliqueShock()):
        from src.io.plot3dio import GridIO
        from src.io.plot3dio import FlowIO
        self.oblique_shock = oblique_shock
        self.nx_max = None
        self.ny_max = None
        self.nz_max = None
        self.points = None
        self.grid = GridIO('dummy')
        self.flow = FlowIO('dummy')
        self.shock_strength = 'weak'
        self.inlet_temperature = None
        self.inlet_density = None
        self.inlet_pressure = None

    def create_grid(self):
        # create a structured grid from -nx_max to nx_max, 0 to ny_max, 0 to nz_max
        # spacing is the grid spacing
        _xx, _yy, _zz = np.meshgrid(np.linspace(-self.nx_max, self.nx_max, 2*self.points),
                                    np.linspace(0, self.ny_max, self.points),
                                    np.linspace(0, self.nz_max, 2), indexing='ij')
        # Create a plot3dio similar object
        # n-blocks - one for the current case
        self.grid.nb = 1
        # ni, nj, nk
        self.grid.ni = np.array([2*self.points], dtype='i4')
        self.grid.nj = np.array([self.points], dtype='i4')
        self.grid.nk = np.array([2], dtype='i4')
        # grd
        # expand dimensions and stack along the last axis
        self.grid.grd = np.stack((_xx[..., None], _yy[..., None], _zz[..., None]), axis=3)
        # grd_min and grd_max - 2d array to match the rest of the code
        self.grid.grd_min = np.array([[-self.nx_max, 0, 0]])
        self.grid.grd_max = np.array([[self.nx_max, self.ny_max, self.nz_max]])

        # compute metrics
        from src.io.plot3dio import GridIO
        GridIO.compute_metrics(self.grid)
        return

    def create_flow(self):
        # Create a vector with density, shock-normal, shock-tangential, zero velocities, and energy
        # Have pre-shock before x=0 and post-shock after
        if self.shock_strength == 'weak':
            self.oblique_shock.shock_angle = self.oblique_shock.shock_angle[0]
            self.oblique_shock.density_ratio = self.oblique_shock.density_ratio[0]
            self.oblique_shock.pressure_ratio = self.oblique_shock.pressure_ratio[0]
            self.oblique_shock.temperature_ratio = self.oblique_shock.temperature_ratio[0]
            self.oblique_shock.mach_ratio = self.oblique_shock.mach_ratio[0]
        elif self.shock_strength == 'strong':
            self.oblique_shock.shock_angle = self.oblique_shock.shock_angle[1]
            self.oblique_shock.density_ratio = self.oblique_shock.density_ratio[1]
            self.oblique_shock.pressure_ratio = self.oblique_shock.pressure_ratio[1]
            self.oblique_shock.temperature_ratio = self.oblique_shock.temperature_ratio[1]
            self.oblique_shock.mach_ratio = self.oblique_shock.mach_ratio[1]
        # Compute inlet flow properties -- pre-shock
        _density = self.inlet_density
        _velocity = self.oblique_shock.mach * np.sqrt(self.oblique_shock.gamma * self.oblique_shock.gas_constant *
                                                      self.inlet_temperature)
        _x_velocity = _velocity * np.sin(np.radians(self.oblique_shock.shock_angle))
        _y_velocity = _velocity * np.cos(np.radians(self.oblique_shock.shock_angle))
        _z_velocity = 0
        _energy = _density * (self.oblique_shock.gamma * self.inlet_temperature / (self.oblique_shock.gamma - 1)
                              + 0.5 * (_x_velocity ** 2 + _y_velocity ** 2 + _z_velocity ** 2))
        # _pre_shock properties
        _pre_shock = np.array([_density, _x_velocity * _density, _y_velocity * _density, _z_velocity * _density,
                               _energy])

        # post-shock properties
        _density_post = _density * self.oblique_shock.density_ratio
        _velocity_post = self.oblique_shock.mach * self.oblique_shock.mach_ratio * np.sqrt(
            self.oblique_shock.gamma * self.oblique_shock.gas_constant
            * self.inlet_temperature * self.oblique_shock.temperature_ratio)
        _x_velocity_post = _velocity_post * np.sin(np.radians(self.oblique_shock.shock_angle
                                                              - self.oblique_shock.deflection))
        _y_velocity_post = _velocity_post * np.cos(np.radians(self.oblique_shock.shock_angle
                                                              - self.oblique_shock.deflection))
        _z_velocity_post = 0
        _energy_post = _density_post * (self.oblique_shock.gamma * self.inlet_temperature *
                                        self.oblique_shock.temperature_ratio / (self.oblique_shock.gamma - 1)
                                        + 0.5 * (_x_velocity_post ** 2 + _y_velocity_post ** 2 + _z_velocity_post ** 2))
        _post_shock = np.array([_density_post, _x_velocity_post * _density_post, _y_velocity_post * _density_post,
                                _z_velocity_post * _density_post, _energy_post])

        # Create a vector with density, shock-normal, shock-tangential, zero velocities, and energy
        # Have pre-shock before x=0 and post-shock after

        # create plot3dio flow similar object
        self.flow.nb = 1
        # ni, nj, nk
        self.flow.ni = np.array([2*self.points], dtype='i4')
        self.flow.nj = np.array([self.points], dtype='i4')
        self.flow.nk = np.array([2], dtype='i4')
        # mach, aoa/alpha, re, t
        self.flow.mach = self.oblique_shock.mach
        self.flow.alpha = 0.0
        self.flow.rey = (_density * _velocity * (self.grid.grd[1, 0, 0, 0, 0] - self.grid.grd[0, 0, 0, 0, 0])) / 1.8e-5
        self.flow.t = 1.0
        # flow
        self.flow.q = np.zeros((self.flow.ni[0], self.flow.nj[0], self.flow.nk[0], 5, self.flow.nb), dtype='f8')
        self.flow.q[:self.points, ...] = _pre_shock[..., None]
        self.flow.q[self.points:, ...] = _post_shock[..., None]
        return





import unittest


class TestImport(unittest.TestCase):
    def test_import(self):
        from src.io import GridIO, FlowIO
        from src.function import Variables
        from src.streamlines import Search, Interpolation, Integration, Streamlines

        grd = GridIO(filename='../data/plate_data/plate.sp.x')
        flw = FlowIO('../data/plate_data/sol-0000010.q')

        import src

        grd = src.io.GridIO(filename='../data/plate_data/plate.sp.x')
        flw = src.io.FlowIO('../data/plate_data/sol-0000010.q')


if __name__ == '__main__':
    unittest.main()
